#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-26 16:17:42 (ywatanabe)"

import re

import matplotlib
import matplotlib.pyplot as plt
import mngs
import pandas as pd


def learning_curve(
    metrics_df,
    keys_to_plot=["loss"],
    title="Title",
    max_n_ticks=4,
    scattersize=3,
    linewidth=1,
    yscale="linear",
):
    COLORS_DICT = {
        "Training": "blue",
        "Validation": "green",
        "Test": "red",
    }

    fig, axes = plt.subplots(len(keys_to_plot), 1, sharex=True, sharey=False)

    axes = axes if len(keys_to_plot) != 1 else [axes]

    axes[-1].set_xlabel("Iteration#")
    fig.text(0.5, 0.95, title, ha="center")

    for i_plt, key_plt in enumerate(keys_to_plot):
        ax = axes[i_plt]
        ax.set_yscale(yscale)
        # ax.set_ylabel(self._rename_if_key_to_plot(key_plt))
        ax.set_ylabel(key_plt)
        ax.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_n_ticks))
        ax.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_n_ticks))

        if re.search("[aA][cC][cC]", key_plt):  # acc, ylim, yticks
            ax.set_ylim(0, 1)
            ax.set_yticks([0, 0.5, 1.0])

        # for step in self.dfs_pivot_i_global.keys():
        for step in COLORS_DICT.keys():

            # if (step == re.search("^[Tt]rain", step)) or (step == re.search("^[Tt]raining", step)):
            if re.match("^[Tt]rain(ing)?", step):  # [REVISED]
                ax.plot(
                    metrics_df[metrics_df.step == step].index,
                    metrics_df[metrics_df.step == step][key_plt],
                    label=step,
                    color=mngs.plt.colors.to_RGBA(
                        COLORS_DICT[step], alpha=0.9
                    ),
                    linewidth=linewidth,
                )
                ax.legend()
                ########################################
                ## Epoch starts points; just in "Training" not to b duplicated
                ########################################
                epoch_starts = abs(
                    metrics_df[metrics_df.step == step]["i_epoch"]
                    - metrics_df[metrics_df.step == step]["i_epoch"].shift(-1)
                )
                indi_global_epoch_starts = [0] + list(
                    epoch_starts[epoch_starts == 1].index
                )

                for i_epoch, i_global_epoch_start in enumerate(
                    indi_global_epoch_starts
                ):
                    ax.axvline(
                        x=i_global_epoch_start,
                        ymin=-1e4,  # ax.get_ylim()[0],
                        ymax=1e4,  # ax.get_ylim()[1],
                        linestyle="--",
                        color=mngs.plt.colors.to_RGBA("gray", alpha=0.1),
                    )
                ########################################

            if (step == "Validation") or (step == "Test"):  # scatter
                ax.scatter(
                    metrics_df[metrics_df.step == step].index,
                    metrics_df[metrics_df.step == step][key_plt],
                    label=step,
                    color=mngs.plt.colors.to_RGBA(
                        COLORS_DICT[step], alpha=0.9
                    ),
                    s=scattersize,
                    alpha=0.9,
                )
                ax.legend()

    return fig


if __name__ == "__main__":
    sdir = "./scripts/train_EEGPT/2024-0126-09:25_bvYp7bIr/"
    metrics = mngs.io.load(sdir + "metrics.csv")
    # del metrics["Unnamed: 0"]
    metrics = metrics.set_index("i_global")
    fig = learning_curve(metrics, yscale="log")

    mngs.io.save(fig, sdir + "learning_curve.png")



#!/usr/bin/env python3


def ax_set_position(
    fig,
    ax,
    x_delta_offset_cm,
    y_delta_offset_cm,
    dragh=False,
    dragv=False,
):

    bbox = ax.get_position()

    ## Calculates delta ratios
    fig_width_inch, fig_height_inch = fig.get_size_inches()

    x_delta_offset_inch = float(x_delta_offset_cm) / 2.54
    y_delta_offset_inch = float(y_delta_offset_cm) / 2.54

    x_delta_offset_ratio = x_delta_offset_inch / fig_width_inch
    y_delta_offset_ratio = y_delta_offset_inch / fig_width_inch

    ## Determines updated bbox position
    left = bbox.x0 + x_delta_offset_ratio
    bottom = bbox.y0 + y_delta_offset_ratio
    width = bbox.x1 - bbox.x0
    height = bbox.y1 - bbox.y0

    if dragh:
        width -= x_delta_offset_ratio

    if dragv:
        height -= y_delta_offset_ratio

    ax.set_position(
        [
            left,
            bottom,
            width,
            height,
        ]
    )

    return ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-24 05:56:27 (ywatanabe)"#!/usr/bin/env python3


import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import skimage
import torch
from obspy.signal.tf_misfit import cwt


def wavelet(data, samp_rate, f_min=100, f_max=None, plot=False):
    """
    Compute the continuous wavelet transform (CWT) of a signal and optionally plot the scalogram.

    Arguments:
        data (numpy.ndarray | torch.Tensor | pandas.DataFrame): The input signal waveform.
        samp_rate (float): The sampling rate of the signal in Hz.
        f_min (float, optional): The minimum frequency of interest in Hz. Defaults to 100.
        f_max (float, optional): The maximum frequency of interest in Hz. Defaults to half the sampling rate.
        plot (bool, optional): If True, plot the scalogram. Defaults to False.

    Returns:
        pandas.DataFrame: A DataFrame containing the absolute values of the CWT coefficients,
                          indexed by frequency in Hz.

    Data Types and Shapes:
        - The input `data` should be a 1D numpy.ndarray, torch.Tensor, or pandas.Series.
        - The output is a pandas.DataFrame with frequencies as the index and time as the columns.

    References:
        - ObsPy's cwt function: https://docs.obspy.org/packages/autogen/obspy.signal.tf_misfit.cwt.html

    Examples:
        # Generate a sample signal
        samp_rate = 1000  # Sampling rate in Hz
        t = np.linspace(0, 1, samp_rate)
        wave = np.sin(2 * np.pi * 50 * t) + 0.5 * np.sin(2 * np.pi * 80 * t)

        # Compute the wavelet transform
        result = wavelet(wave, samp_rate, f_min=10, f_max=500, plot=True)
        print(result)
    """
    # Convert input data to a NumPy array if it is not already one
    if isinstance(data, torch.Tensor):
        wave = data.numpy()
    elif isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
        wave = data.values.squeeze()
    else:
        wave = data

    # Ensure the input wave is a 1D array
    if wave.ndim != 1:
        raise ValueError("Input data must be a 1D array or Series.")

    dt = 1.0 / samp_rate
    npts = len(wave)
    t = np.linspace(0, dt * npts, npts)
    if f_min is None:
        f_min = 0.1
    if f_max is None:
        f_max = int(samp_rate / 2)
    scalogram = cwt(wave, dt, 8, f_min, f_max)

    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        x, y = np.meshgrid(
            t,
            np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0]),
        )

        ax.pcolormesh(x, y, np.abs(scalogram), shading="auto", cmap="viridis")
        ax.set_xlabel("Time [s]")
        ax.set_ylabel("Frequency [Hz]")
        ax.set_yscale("log")
        ax.set_ylim(f_min, f_max)
        plt.show()

    Hz = pd.Series(
        np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0])
    )
    df = pd.DataFrame(np.abs(scalogram), columns=t)
    df["Hz"] = Hz
    df.set_index("Hz", inplace=True)

    return df


# def wavelet(*args, **kwargs):
#     """
#     Deprecated: Use 'compute_wavelet_transform' instead.
#     """
#     warnings.warn(
#         "'wavelet' is deprecated and will be removed in a future version. "
#         "Use 'compute_wavelet_transform' instead.",
#         DeprecationWarning,
#     )
#     return compute_wavelet_transform(*args, **kwargs)


def to_z(x, i_dim=-1):
    """
    Normalize the input data along the specified time dimension.

    The function subtracts the mean and divides by the standard deviation along
    the specified axis for numpy.ndarray and torch.Tensor, or along the columns
    for pandas.DataFrame.

    Args:
        x (numpy.ndarray or torch.Tensor or pandas.DataFrame): The input data to normalize.
        i_dim (int, optional): The dimension along which to normalize the data.
                                  Defaults to -1 (the last dimension).

    Returns:
        numpy.ndarray or torch.Tensor or pandas.DataFrame: The normalized data with the same type as the input.

    Examples:
        numpy_array = np.random.rand(16, 160, 1000)
        normalized_array = normalize_time(numpy_array, i_dim=2)

        torch_tensor = torch.randn(16, 160, 1000)
        normalized_tensor = normalize_time(torch_tensor, i_dim=2)

        pandas_dataframe = pd.DataFrame(np.random.rand(160, 1000))
        normalized_dataframe = normalize_time(pandas_dataframe)  # i_dim is ignored for DataFrame
    """
    if isinstance(x, torch.Tensor):
        mean = x.mean(dim=i_dim, keepdim=True)
        std = x.std(dim=i_dim, keepdim=True)
        return (x - mean) / std

    elif isinstance(x, np.ndarray):
        mean = x.mean(axis=i_dim, keepdims=True)
        std = x.std(axis=i_dim, keepdims=True)
        return (x - mean) / std

    elif isinstance(x, pd.DataFrame):
        if i_dim != -1 and i_dim != 1:
            raise ValueError(
                "For pandas.DataFrame, i_dim must be -1 or 1 (columns)."
            )
        mean = x.mean(axis=1).values.reshape(-1, 1)
        std = x.std(axis=1).values.reshape(-1, 1)
        return (x - mean) / std

    else:
        raise TypeError(
            "Input must be a numpy.ndarray, torch.Tensor, or pandas.DataFrame"
        )


def wavelet_np(wave, samp_rate, f_min=100, f_max=None, plot=False, title=None):
    """
    Perform a continuous wavelet transform on a signal and optionally plot the scalogram.

    Parameters:
        wave (numpy.ndarray): The input signal. Shape should be (n_samples,).
        samp_rate (float): The sampling rate of the signal in Hz.
        f_min (float): The minimum frequency for the wavelet transform. Defaults to 100 Hz.
        f_max (float, optional): The maximum frequency for the wavelet transform. Defaults to half the sampling rate.
        plot (bool): If True, plot the scalogram. Defaults to False.
        title (str, optional): The title for the plot if plot is True.

    Returns:
        pandas.DataFrame: A DataFrame containing the wavelet transform scalogram with frequencies as the index.

    Data Types:
        Input:
            wave: numpy.ndarray (typically float32 or float64)
            samp_rate: float
            f_min: float
            f_max: float or None
            plot: bool
            title: str or None
        Output:
            pandas.DataFrame

    Data Shapes:
        Input:
            wave: (n_samples,)
        Output:
            DataFrame: (n_frequencies, n_samples)

    References:
        - Continuous wavelet transform: https://en.wikipedia.org/wiki/Continuous_wavelet_transform
        - scipy.signal.cwt: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.cwt.html
        - obspy.imaging.cm.obspy_sequential: https://docs.obspy.org/packages/autogen/obspy.imaging.cm.obspy_sequential.html
        - pandas.DataFrame: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html
        - matplotlib.pyplot: https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.html

    Raises:
        ValueError: If the input signal is not a 1D numpy.ndarray.
    """

    dt = 1.0 / samp_rate
    npts = len(wave)
    t = np.linspace(0, dt * npts, npts)
    if f_min is None:
        f_min = 0.1
    if f_max is None:
        f_max = int(samp_rate / 2)
        scalogram = cwt(wave, dt, 8, f_min, f_max)

    if plot:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        x, y = np.meshgrid(
            t,
            np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0]),
        )

        ax.pcolormesh(x, y, np.abs(scalogram), cmap=obspy_sequential)
        ax.set_xlabel("Time [s]")
        ax.set_ylabel("Frequency [Hz]")
        ax.set_yscale("log")
        ax.set_title(title)
        ax.set_ylim(f_min, f_max)
        fig.show()

    Hz = pd.Series(
        np.logspace(np.log10(f_min), np.log10(f_max), scalogram.shape[0])
    )
    df = pd.DataFrame(np.abs(scalogram))
    df["Hz"] = Hz
    df.set_index("Hz", inplace=True)

    return df


def psd(signals_2d, samp_rate, normalize=True):
    """
    Calculate the Power Spectral Density (PSD) of each signal in a 2D array, tensor, or DataFrame.

    Arguments:
        signals_2d (numpy.ndarray, torch.Tensor, or pandas.DataFrame): A 2D collection containing multiple signals,
                                                                       where each row represents a signal.
        samp_rate (float): The sampling rate of the signals.
        normalize (bool, optional): If True, normalize the PSD by the sum of powers for each signal.
                                    Defaults to True.

    Returns:
        pandas.DataFrame: A DataFrame containing the PSD for each signal. The columns represent the
                          frequencies, and the rows represent the individual signals.

    Data Types:
        Input can be either numpy.ndarray, torch.Tensor, or pandas.DataFrame. Output is always a pandas.DataFrame.

    Data Shapes:
        - signals_2d: (n_signals, signal_length)
        - Output DataFrame: (n_signals, n_frequencies)

    References:
        - NumPy documentation for FFT: https://numpy.org/doc/stable/reference/routines.fft.html
        - PyTorch documentation for FFT: https://pytorch.org/docs/stable/fft.html
        - pandas.DataFrame documentation: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html
    """
    # Convert input to NumPy array if it is a pandas DataFrame or PyTorch tensor
    if isinstance(signals_2d, pd.DataFrame):
        signals_2d = signals_2d.values
    elif isinstance(signals_2d, torch.Tensor):
        signals_2d = signals_2d.numpy()

    # Calculate the FFT for each signal
    fft_results = np.fft.rfft(signals_2d, axis=1)

    # Calculate the PSD by taking the absolute square of the FFT results
    psd = np.abs(fft_results) ** 2

    # Normalize the PSD if requested
    if normalize:
        psd /= np.sum(psd, axis=1, keepdims=True)

    # Calculate the frequency bins
    freqs = np.fft.rfftfreq(signals_2d.shape[1], d=1 / samp_rate)

    # Create a DataFrame for the PSD with frequency labels
    psd_df = pd.DataFrame(data=psd, columns=freqs)

    return psd_df


def arr2skdf(data):
    """
    Convert a 3D numpy array into a DataFrame suitable for sktime.

    Parameters:
    data (numpy.ndarray): A 3D numpy array with shape (n_samples, n_channels, seq_len)

    Returns:
    pandas.DataFrame: A DataFrame in sktime format
    """
    if len(data.shape) != 3:
        raise ValueError("Input data must be a 3D array")

    n_samples, seq_len, n_channels = data.shape

    # Initialize an empty DataFrame for sktime format
    sktime_df = pd.DataFrame(index=range(n_samples), columns=["dim_0"])

    # Iterate over each sample
    for i in range(n_samples):
        # Combine all channels into a single cell
        combined_series = pd.Series(
            {
                f"channel_{j}": pd.Series(data[i, :, j])
                for j in range(n_channels)
            }
        )
        sktime_df.iloc[i, 0] = combined_series

    return sktime_df


def crop(data, window_size_pts, overlap_factor=1):
    assert data.ndim == 2

    cropped = skimage.util.view_as_windows(
        data,
        (len(data), window_size_pts),
        int(window_size_pts / overlap_factor),
    )

    if cropped.ndim != 3:
        cropped = cropped.squeeze(0)
    assert cropped.ndim == 3

    return cropped


# def normalize_time(x):
#     if type(x) == torch.Tensor:
#         return (x - x.mean(dim=-1, keepdims=True)) \
#             / x.std(dim=-1, keepdims=True)
#     if type(x) == np.ndarray:
#         return (x - x.mean(axis=-1, keepdims=True)) \
#             / x.std(axis=-1, keepdims=True)

if __name__ == "__main__":
    x = 100 * np.random.rand(16, 160, 1000)
    print(_normalize_time(x))

    x = torch.randn(16, 160, 1000)
    print(_normalize_time(x))

    x = torch.randn(16, 160, 1000).cuda()
    print(_normalize_time(x))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-20 13:58:43 (ywatanabe)"#!/usr/bin/env python3

import numpy as np
import torch


def demo_sig_np(
    T_sec=50.0, fs=250, freqs_hz=[30, 60, 100, 200, 1000], n_chs=19
):
    """
    Prepare a demo dataset composed of multiple channels of sinusoidal waves.

    Parameters:
        T_sec (float): The duration of the signal in seconds. Defaults to 50.0 seconds.
        fs (int): The sampling frequency of the signal in samples per second. Defaults to 250 Hz.
        freqs_hz (list of float): A list of frequencies in Hz for the sinusoidal components. Defaults to [30, 60, 100, 200, 1000].
        n_chs (int): The number of channels to generate. Defaults to 19.

    Returns:
        numpy.ndarray: A 2D array representing the generated multi-channel signal, with shape (n_chs, T_sec * fs).

    Data Types:
        Input:
            T_sec: float
            fs: int
            freqs_hz: list of float
            n_chs: int
        Output:
            numpy.ndarray (typically float64)

    Data Shapes:
        Input:
            freqs_hz: (n_frequencies,)
        Output:
            numpy.ndarray: (n_chs, n_samples)

    References:
        - Sinusoidal wave: https://en.wikipedia.org/wiki/Sine_wave
        - NumPy array: https://numpy.org/doc/stable/reference/generated/numpy.array.html

    Examples:
        # Prepare a demo dataset with a 10-second duration, sampled at 500 Hz, with frequencies 50, 100, and 150 Hz, and 16 channels
        demo_data = prepair_demo_data(T_sec=10.0, fs=500, freqs_hz=[50, 100, 150], n_chs=16)
    """
    data = np.array(
        [
            summarize_sinusoidal_waves(T_sec=T_sec, fs=fs, freqs_hz=freqs_hz)
            for _ in range(n_chs)
        ]
    )
    return data


def demo_sig_torch(
    batch_size=64, n_chs=19, samp_rate=1000, len_sec=10, freqs_hz=[2, 5, 10]
):
    time = torch.arange(0, len_sec, 1 / samp_rate)
    sig = torch.vstack(
        [torch.sin(f * 2 * torch.pi * time) for f in freqs_hz]
    ).sum(dim=0)
    sig = sig.unsqueeze(0).unsqueeze(0).repeat(batch_size, n_chs, 1)
    return sig


def summarize_sinusoidal_waves(
    T_sec=50.0, fs=250, freqs_hz=[30, 60, 100, 200, 1000]
):
    """
    Generate a signal composed of a sum of sinusoidal waves with random amplitudes and phases.

    Parameters:
        T_sec (float): The duration of the signal in seconds. Defaults to 50.0 seconds.
        fs (int): The sampling frequency of the signal in samples per second. Defaults to 250 Hz.
        freqs_hz (list of float): A list of frequencies in Hz for the sinusoidal components. Defaults to [30, 60, 100, 200, 1000].

    Returns:
        numpy.ndarray: The generated signal, which is a 1D array of length `T_sec * fs`.

    Data Types:
        Input:
            T_sec: float
            fs: int
            freqs_hz: list of float
        Output:
            numpy.ndarray (typically float64)

    Data Shapes:
        Input:
            freqs_hz: (n_frequencies,)
        Output:
            numpy.ndarray: (n_samples,)

    References:
        - Sinusoidal wave: https://en.wikipedia.org/wiki/Sine_wave
        - NumPy linspace: https://numpy.org/doc/stable/reference/generated/numpy.linspace.html
        - NumPy random.rand: https://numpy.org/doc/stable/reference/random/generated/numpy.random.rand.html

    Examples:
        # Generate a 10-second signal sampled at 500 Hz with frequencies 50, 100, and 150 Hz
        signal = summarize_sinusoidal_waves(T_sec=10.0, fs=500, freqs_hz=[50, 100, 150])
    """

    n = int(T_sec * fs)
    t = np.linspace(0, T_sec, n, endpoint=False)
    summed = np.array(
        [
            np.random.rand()
            * np.sin((f_hz * t + np.random.rand()) * (2 * np.pi))
            for f_hz in freqs_hz
        ]
    ).sum(axis=0)
    return summed



#!/usr/bin/env python3
# Time-stamp: "2024-01-28 09:37:16 (ywatanabe)"

from . import dsp
from . import general
from . import general as gen
from . import io, linalg, ml, nn, plt, resource, stats
from .general.debug import *

__copyright__ = "Copyright (C) 2021 Yusuke Watanabe"
__version__ = "0.10.1"
__license__ = "GPL3.0"
__author__ = "ywatanabe1989"
__author_email__ = "ywata1989@gmail.com"
__url__ = "https://github.com/ywatanabe1989/mngs"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-07-10 19:26:42 (ywatanabe)"

def ax_fill_between(ax, xx, mean, std, label, alpha=.1):
    ax.plot(xx, mean, label=label, alpha=alpha)
    ax.fill_between(xx, mean-std, mean+std, alpha=alpha)
    return ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-23 20:15:30 (ywatanabe)"

import mne


def to_dig_montage(channel_names):
    # Load the standard 10-20 montage
    standard_montage = mne.channels.make_standard_montage("standard_1020")

    # Get the positions of the electrodes in the standard montage
    positions = standard_montage.get_positions()

    # Filter the positions to only include the desired channels
    custom_ch_pos = {
        ch: positions["ch_pos"][ch]
        for ch in channel_names
        if ch in positions["ch_pos"]
    }

    # Create a custom DigMontage
    custom_montage = mne.channels.make_dig_montage(
        ch_pos=custom_ch_pos,
        nasion=positions["nasion"],
        lpa=positions["lpa"],
        rpa=positions["rpa"],
    )

    return custom_montage



#!/usr/bin/env python3

from .load import (get_data_path_from_a_package, load, load_study_rdb,
                   load_yaml_as_an_optuna_dict)
from .path import (find_the_git_root_dir, get_this_fpath, mk_spath,
                   split_fpath, touch, find)
from .save import (is_listed_X, save, save_listed_dfs_as_csv,
                   save_listed_scalars_as_csv,
                   save_optuna_study_as_csv_and_pngs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-04-27 11:04:12 (ywatanabe)"

#!/usr/bin/env python3

import itertools
import re
import sys
import warnings
from copy import deepcopy

import bids
import mngs
import numpy as np
import pandas as pd
from bids import BIDSLayout
from mne.io import read_raw_edf
from natsort import natsorted
try:
    from pandas.errors import SettingWithCopyWarning
except:
    from pandas.core.common import SettingWithCopyWarning

# from pandas.core.common import SettingWithCopyWarning
# from pandas.errors import SettingWithCopyWarning
from scipy import stats
from tqdm import tqdm

warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)



def load_BIDS(BIDS_ROOT, channels, verbose=False):
    layout = BIDSLayout(BIDS_ROOT)
    entities = layout.get_entities()

    if verbose:
        layout_keys_to_print = [
            "InstitutionName",
            "PowerLineFrequency",
            "SamplingFrequency",
            "TaskName",
            "extension",
            "run",
            "session",
            "subject",
            "suffix",
            "task",
        ]

        for k in layout_keys_to_print:
            print(f"\nunique {k}s:\n{natsorted(entities[k].unique())}\n")
            sleep(1)

    # Probably, only one instituion is allowed for a BIDS dataset.
    subj_uq = natsorted(entities["subject"].unique())
    task_uq = natsorted(entities["task"].unique())
    datatype_uq = natsorted(entities["datatype"].unique())
    session_uq = natsorted(entities["session"].unique())
    run_uq = natsorted(entities["run"].unique())

    data_all = []
    for i_subj, subj in tqdm(enumerate(subj_uq)):

        try:
            # Session
            session_bids = layout.get(subject=subj, suffix="sessions")
            assert len(session_bids) == 1
            session_bids = session_bids[0]

            for task in task_uq:
                for datatype in datatype_uq:
                    for session in session_uq:
                        for run in run_uq:

                            ## EDF, signal
                            edf_bids = layout.get(
                                subject=subj,
                                session=session,
                                run=run,
                                datatype=datatype,
                                extension=".edf",
                            )
                            assert len(edf_bids) == 1
                            edf_bids = edf_bids[0]

                            ## Event
                            event_bids = layout.get(
                                subject=subj,
                                session=session,
                                run=run,
                                datatype=datatype,
                                suffix="events",
                                extension=".tsv",
                            )
                            assert len(event_bids) == 1
                            event_bids = event_bids[0]

                            ## eeg metadata from json
                            metadata_bids = layout.get(
                                subject=subj,
                                session=session,
                                run=run,
                                datatype=datatype,
                                suffix="eeg",
                                extension=".json",
                            )
                            assert len(metadata_bids) == 1
                            metadata_bids = metadata_bids[0]

                            ## Load data of a run
                            data_run = _load_a_run(
                                session_bids,
                                edf_bids,
                                event_bids,
                                metadata_bids,
                                channels,
                            )

                            ## Buffering
                            data_all.append(data_run)

        except Exception as e:
            print(f"\n{subj}:\n{e}\n")

    data_all_df = pd.concat(data_all).reset_index()
    del data_all_df["index"]

    return data_all_df


def _load_a_run(
    session_bids, edf_bids, event_bids, metadata_bids, channels, verbose=False
):
    event_df = event_bids.get_df()
    event_df["offset"] = event_df["onset"] + event_df["duration"]
    metadata_dict = metadata_bids.get_dict()

    df = event_df.copy()

    df = df[df["trial_type"] == "NoiseFreeSection"]

    eeg_data = _load_an_edf(
        edf_bids,
        channels=channels,
        starts=df["onset"],
        ends=df["offset"],
    )  # load_params["channels"], # fixme

    df["eeg"] = eeg_data

    df = pd.DataFrame(df[["eeg", "trial_type"]])

    ## Labels
    session_df = session_bids.get_df()
    for k, v in session_df.items():
        assert v.shape == (1,)
        df[k] = v[0]
    del df["session_id"]

    entities = edf_bids.get_entities()
    df["subject"] = entities["subject"]
    df["run"] = entities["run"]
    df["session"] = entities["session"]
    df["task"] = entities["task"]
    df["samp_rate"] = metadata_dict["SamplingFrequency"]
    df["channels"] = [channels for _ in range(len(df))]

    return df


def _load_an_edf(bf_edf, channels, starts=[0], ends=[None]):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        mne_raw = read_raw_edf(bf_edf, verbose=False)

    sfreq = mne_raw.info["sfreq"]
    # channels = mne_raw.info["ch_names"]

    data = [
        mne_raw.get_data(picks=channels, start=int(start * sfreq), stop=int(end * sfreq))
        for start, end in zip(starts, ends)
    ]
    # data = [
    #     mne_raw.get_data(
    #         picks=channels, start=int(start * sfreq), stop=int(end * sfreq)
    #     )
    #     for start, end in zip(starts, ends)
    # ]

    return data


if __name__ == "__main__":
    import mngs

    EEG_CHANNELS = [
        "FP1",
        "F3",
        "C3",
        "P3",
        "O1",
        "FP2",
        "F4",
        "C4",
        "P4",
        "O2",
        "F7",
        "T7",
        "P7",
        "F8",
        "T8",
        "P8",
        "Fz",
        "Cz",
        "Pz",
        "A1",
        "A2",        
    ]

    BIDS_FPATH = "YOUR_BIDS_FPATH"
    BIDS_FPATH = "./data/DEM/BIDS_Osaka"
    data_all = load_BIDS(BIDS_FPATH, channels=EEG_CHANNELS)
    mngs.io.save(data_all, "./res/DEM/dataset.pkl")

    data_all.columns
    # Index(['eeg', 'trial_type', 'disease_type', 'cognitive_level', 'pet',
    #  'subject', 'run', 'session', 'task', 'samp_rate', 'channels'],
    # dtype='object')



#!/usr/bin/env python3

from . import colors
from ._ax_extend import ax_extend
from ._ax_scientific_notation import ax_scientific_notation
from ._ax_set_n_ticks import ax_set_n_ticks
from ._ax_set_position import ax_set_position
from ._ax_set_size import ax_set_size
from ._configure_mpl import configure_mpl
from ._draw_a_cube import draw_a_cube
from ._get_RGBA_from_colormap import get_RGBA_from_colormap
from .annotated_heatmap import annotated_heatmap
from ._mk_patches import mk_patches
from ._mk_colorbar import mk_colorbar
from ._add_hue import add_hue
from ._ax_fill_between import ax_fill_between
from ._ax_circular_hist import ax_circular_hist



#!/usr/bin/env python

import numpy as np
import torch  # 1.7.1
import torch.nn as nn
from torch.fft import fft, ifft


class BaseHilbertTransformerTorch(nn.Module):
    """
    Base class for Hilbert transformation on a given axis of the input tensor.

    Attributes:
        axis (int): The axis along which to compute the Hilbert transform.
        n (int or None): The number of points for the FFT computation. If None, the number of points is taken from the input tensor's size along the axis.

    References:
        - MONAI implementation of Hilbert transform: https://github.com/Project-MONAI/MONAI/blob/master/monai/networks/layers/simplelayers.py
    """

    def __init__(self, axis=2, n=None):
        """
        Initialize the BaseHilbertTransformerTorch.

        Args:
            axis (int): The axis along which to compute the Hilbert transform.
            n (int or None): The number of points for the FFT computation. If None, the number of points is taken from the input tensor's size along the axis.
        """
        super().__init__()
        self.axis = axis
        self.n = n

    def transform(self, x):
        """
        Apply the Hilbert transform to the input tensor along the specified axis.

        Args:
            x (torch.Tensor): The input tensor to transform. Must be real and in shape [Batch, chns, spatial1, spatial2, ...].

        Returns:
            torch.Tensor: The analytical signal of `x`, transformed along the axis specified in `self.axis` using FFT of size `self.n`. The absolute value of the result relates to the envelope of `x` along `self.axis`.
        """
        n = x.shape[self.axis] if self.n is None else self.n

        # Create frequency axis
        f = torch.cat(
            [
                torch.arange(0, (n - 1) // 2 + 1, device=x.device) / float(n),
                torch.arange(-(n // 2), 0, device=x.device) / float(n),
            ]
        )

        xf = fft(x, n=n, dim=self.axis)

        # Create step function
        u = torch.heaviside(f, torch.tensor([0.5], device=f.device))
        u = u.to(dtype=x.dtype, device=x.device)
        new_dims_before = self.axis
        new_dims_after = len(xf.shape) - self.axis - 1
        for _ in range(new_dims_before):
            u = u.unsqueeze(0)
        for _ in range(new_dims_after):
            u = u.unsqueeze(-1)

        transformed = ifft(xf * 2 * u, dim=self.axis)

        return transformed


# class BaseHilbertTransformerTorch(nn.Module):
#     # https://github.com/Project-MONAI/MONAI/blob/master/monai/networks/layers/simplelayers.py
#     def __init__(self, axis=2, n=None):
#         super().__init__()
#         self.axis = axis
#         self.n = n

#     def transform(self, x):
#         """
#         Args:
#             x: Tensor or array-like to transform.
#                Must be real and in shape ``[Batch, chns, spatial1, spatial2, ...]``.
#         Returns:
#             torch.Tensor: Analytical signal of ``x``,
#                           transformed along axis specified in ``self.axis`` using
#                           FFT of size ``self.N``.
#                           The absolute value of ``x_ht`` relates to the envelope of ``x``
#                           along axis ``self.axis``.
#         """

#         n = x.shape[self.axis] if self.n is None else self.n

#         # Create frequency axis
#         f = torch.cat(
#             [
#                 torch.true_divide(
#                     torch.arange(0, (n - 1) // 2 + 1, device=x.device), float(n)
#                 ),
#                 torch.true_divide(
#                     torch.arange(-(n // 2), 0, device=x.device), float(n)
#                 ),
#             ]
#         )

#         xf = fft(x, n=n, dim=self.axis)

#         # Create step functionb
#         u = torch.heaviside(f, torch.tensor([0.5], device=f.device))
#         u = torch.as_tensor(u, dtype=x.dtype, device=u.device)
#         new_dims_before = self.axis
#         new_dims_after = len(xf.shape) - self.axis - 1
#         for _ in range(new_dims_before):
#             u.unsqueeze_(0)
#         for _ in range(new_dims_after):
#             u.unsqueeze_(-1)

#         transformed = ifft(xf * 2 * u, dim=self.axis)

#         return transformed


class HilbertTransformerTorch(BaseHilbertTransformerTorch):
    """
    Hilbert transformer module that computes the analytical signal of the input tensor and extracts its amplitude and phase components.

    Inherits from BaseHilbertTransformerTorchTorch.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the HilbertTransformerTorch.

        Args:
            *args: Variable length argument list to pass to the BaseHilbertTransformerTorchTorch.
            **kwargs: Arbitrary keyword arguments to pass to the BaseHilbertTransformerTorchTorch.
        """
        super().__init__(*args, **kwargs)
        self.hilbert_transform = self.transform
        self.register_buffer("pi", torch.tensor(np.pi))

    def forward(self, sig):
        """
        Apply the Hilbert transform to the input signal and extract its amplitude and phase components.

        Args:
            sig (torch.Tensor): The input signal tensor.

        Returns:
            torch.Tensor: A tensor containing the phase and amplitude components of the input signal's analytical signal.
        """
        sig_comp = self.hilbert_transform(sig)

        pha = self._calc_Arg(sig_comp)
        amp = sig_comp.abs()

        out = torch.cat(
            [
                pha.unsqueeze(-1),
                amp.unsqueeze(-1),
            ],
            dim=-1,
        )

        return out

    def _calc_Arg(self, comp):
        """
        Calculates the argument (phase) of complex numbers in the range (-pi, pi].

        Args:
            comp (torch.Tensor): The complex tensor for which to calculate the argument.

        Returns:
            torch.Tensor: The argument (phase) of the complex tensor.
        """
        x, y = comp.real, comp.imag
        Arg = torch.zeros_like(x).type_as(x)

        # Conditions and corresponding phase calculations
        c1 = x > 0
        Arg[c1] = torch.atan(y / x)[c1]

        c2 = (x < 0) & (y >= 0)
        Arg[c2] = torch.atan(y / x)[c2] + self.pi

        c3 = (x < 0) & (y < 0)
        Arg[c3] = torch.atan(y / x)[c3] - self.pi

        c4 = (x == 0) & (y > 0)
        Arg[c4] = self.pi / 2.0

        c5 = (x == 0) & (y < 0)
        Arg[c5] = -self.pi / 2.0

        c6 = (x == 0) & (y == 0)
        Arg[c6] = 0.0

        return Arg


# class HilbertTransformerTorch(BaseHilbertTransformerTorchTorch):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.hilbert_transform = self.transform
#         self.register_buffer("pi", torch.tensor(np.pi))

#     def forward(self, sig):
#         sig_comp = self.hilbert_transform(sig)

#         pha = self._calc_Arg(sig_comp)
#         amp = sig_comp.abs()

#         out = torch.cat(
#             [
#                 pha.unsqueeze(-1),
#                 amp.unsqueeze(-1),
#             ],
#             dim=-1,
#         )

#         return out

#     def _calc_Arg(self, comp):
#         """Calculates argument of complex numbers in (-pi, pi] space.
#         Although torch.angle() does not have been implemented the derivative function,
#         this function seems to work.
#         """
#         x, y = comp.real, comp.imag
#         Arg = torch.zeros_like(x).type_as(x)

#         c1 = x > 0  # condition #1
#         Arg[c1] += torch.atan(y / x)[c1]

#         c2 = (x < 0) * (y >= 0)
#         Arg[c2] += torch.atan(y / x)[c2] + self.pi

#         c3 = (x < 0) * (y < 0)
#         Arg[c3] += torch.atan(y / x)[c3] - self.pi

#         c4 = (x == 0) * (y > 0)
#         Arg[c4] += self.pi / 2.0

#         c5 = (x == 0) * (y < 0)
#         Arg[c5] += -self.pi / 2.0

#         c6 = (x == 0) * (y == 0)
#         Arg[c6] += 0.0

#         return Arg


def _unwrap(x):
    """
    Unwrap a phase tensor by changing absolute jumps greater than `pi` to their 2*pi complement.

    Args:
        x (torch.Tensor): The input phase tensor.

    Returns:
        torch.Tensor: The unwrapped phase tensor.
    """
    pi = torch.tensor(np.pi)
    y = x % (2 * pi)
    return torch.where(y > pi, y - 2 * pi, y)


# def _unwrap(x):
#     pi = torch.tensor(np.pi)
#     y = x % (2 * pi)
#     return torch.where(y > pi, 2 * pi - y, y)


def test_hilbert_torch(sig, axis=1, test_layer=True):
    """
    Test function for the HilbertTransformerTorch.

    Args:
        sig (torch.Tensor): The input signal tensor.
        axis (int): The axis along which to apply the Hilbert transform.
        test_layer (bool): If True, use the HilbertTransformerTorch as a layer; otherwise, use the transform method directly.

    Returns:
        tuple: A tuple containing the phase and amplitude components of the input signal's analytical signal.
    """
    if test_layer:
        # Hilbert Layer
        hilbert_layer = HilbertTransformerTorch(axis=axis).cuda()
        sig_comp = hilbert_layer(sig)
    else:
        # Hilbert Transformation
        hilbert = HilbertTransformerTorch(axis=axis)
        sig_comp = hilbert.transform(sig)

    print(sig.shape)
    print(sig_comp.shape)

    print(sig.dtype)
    print(sig_comp.dtype)

    print(sig_comp)

    pha, amp = sig_comp[..., 0], sig_comp[..., 1]
    return pha, amp


# def test_hilbert_torch(sig, axis=1, test_layer=True):

#     if test_layer:
#         ## Hilbert Layer
#         hilbert_layer = HilbertTransformerTorch(axis=axis).cuda()
#         sig_comp = hilbert_layer(sig)

#     else:
#         ## Hilbert Transformation
#         hilbert = HilbertTransformerTorch(axis=1)
#         sig_comp = hilbert.transform(sig)

#     print(sig.shape)
#     print(sig_comp.shape)

#     print(sig.dtype)
#     print(sig_comp.dtype)

#     print(sig_comp)

#     # ## Extract Amplitude and Phase signals
#     # amp = sig_comp.abs()
#     # pha = _unwrap(sig_comp.angle())

#     # instantaneous_freq = (np.diff(pha.cpu()) / (2.0*np.pi) * fs)

#     # ## Plot
#     # fig, ax = plt.subplots(3, 1)
#     # ax[0].set_title('GPU Hilbert Transformation')

#     # ax[0].plot(t, sig[0].cpu(), label='original signal')
#     # ax[0].plot(t, amp[0].cpu(), label='envelope')
#     # ax[0].set_xlabel("time in seconds")
#     # ax[0].legend()

#     # ax[1].plot(t, pha[0].cpu(), label='phase')
#     # ax[1].set_xlabel("time in seconds")
#     # ax[1].legend()

#     # ax[2].plot(t[1:], instantaneous_freq[0], label='instantenous frequency')
#     # ax[2].set_xlabel("time in seconds")
#     # ax[2].set_ylim(0.0, 120.0)
#     # ax[2].legend()

#     # fig.show()

#     pha, amp = sig_comp[..., 0], sig_comp[..., 1]
#     return pha, amp


def mk_sig():
    """
    Create a batch of demo signals using a chirp function.

    Returns:
        torch.Tensor: A batch of demo signals as a CUDA tensor.
    """
    from scipy.signal import chirp

    def _mk_sig():
        sig = chirp(t, 20.0, t[-1], 100.0)
        sig *= 1.0 + 0.5 * np.sin(2.0 * np.pi * 3.0 * t)
        return sig

    sig = np.array([_mk_sig() for _ in range(batch_size)]).astype(np.float32)
    sig = torch.tensor(sig).cuda()
    return sig


# def mk_sig():
#     from scipy.signal import chirp

#     def _mk_sig():
#         sig = chirp(t, 20.0, t[-1], 100.0)
#         sig *= 1.0 + 0.5 * np.sin(2.0 * np.pi * 3.0 * t)
#         return sig

#     sig = np.array([_mk_sig() for _ in range(batch_size)]).astype(np.float32)
#     sig = torch.tensor(sig).cuda()
#     return sig


if __name__ == "__main__":
    # import numpy as np
    import matplotlib
    import matplotlib.pyplot as plt
    from scipy.signal import chirp

    ## Parameters
    duration = 1.0
    fs = 400.0
    samples = int(fs * duration)
    t = np.arange(samples) / fs
    batch_size = 64

    ## Create demo signal
    sig = mk_sig()

    ## Test Code
    pha, amp = test_hilbert_torch(sig, axis=1, test_layer=True)
    # ## EOF



#!/usr/bin/env python3

# from .path import find_the_git_root_dir, get_this_fpath, mk_spath, split_fpath
# from .load import get_data_path_from_a_package, load
# from .save import is_listed_X, save, save_listed_dfs_as_csv, save_listed_scalars_as_csv
from ..io.__init__ import *
from ._close import close
from ._start import start
from .calc import to_z
from .converters import (
    KeepPD,
    my2array,
    my2tensor,
    three2two_dim,
    two2three_dim,
)
from .cuda_collect_env import main as cuda_collect_env
from .debug import *
from .email import notify, send_gmail
from .latex import add_hat_in_the_latex_style, to_the_latex_style
from .mat2py import *
from .misc import (
    _return_counting_process,
    connect_nums,
    connect_strs,
    copy_files,
    copy_the_file,
    decapitalize,
    describe,
    fmt_size,
    grep,
    is_defined_global,
    is_defined_local,
    is_later_or_equal,
    is_nan,
    isclose,
    listed_dict,
    merge_dicts_wo_overlaps,
    partial_at,
    pop_keys,
    search,
    squeeze_spaces,
    suppress_output,
    take_the_closest,
    wait_key,
)
from .pandas import (
    col_to_last,
    col_to_top,
    force_dataframe,
    ignore_SettingWithCopyWarning,
    merge_columns,
)
from .repro import *
from .TimeStamper import *
from .torch import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-20 23:36:39 (ywatanabe)"

import numpy as np
import pandas as pd
import torch

def common_average(sig_2D):
    """
    Normalize a 2D signal by subtracting the mean and dividing by the standard deviation.

    This function computes the common average referencing of a 2D signal array, where
    the mean is subtracted from each element and the result is divided by the standard
    deviation of the entire array. This is a common preprocessing step in signal processing
    to standardize the signal.

    Parameters
    ----------
    sig_2D : ndarray
        A 2D NumPy array of signal values. The array should not be empty.

    Returns
    -------
    ndarray
        The normalized 2D signal array, with the same shape as the input `sig_2D`.

    Notes
    -----
    The function does not handle cases where the standard deviation is zero (i.e., all
    elements in `sig_2D` are the same). In such cases, the function will raise a
    `RuntimeWarning` due to division by zero.

    Examples
    --------
    >>> import numpy as np
    >>> sig_2D = np.array([[1, 2, 3], [4, 5, 6]])
    >>> common_average(sig_2D)
    array([[-1.46385011, -0.87831007, -0.29277002],
           [ 0.29277002,  0.87831007,  1.46385011]])

    References
    ----------
    For more information on common average referencing and its applications in signal
    processing, see:
    Nunez, P. L., & Srinivasan, R. (2006). Electric fields of the brain: The
    neurophysics of EEG. Oxford University Press, USA.

    """
    return (sig_2D - sig_2D.mean()) / sig_2D.std()


def subtract_from_random_column(data, random_state=42, rename=False):
    np.random.seed(random_state)

    if torch.is_tensor(data):
        data = data.numpy()
    
    if isinstance(data, np.ndarray):
        data = pd.DataFrame(data)

    cols_orig = data.columns.tolist()
    cols_shuffled = np.random.permutation(cols_orig)
    cols_diff = [f"{co} - {cs}" for co, cs in zip(cols_orig, cols_shuffled)]

    diff_df = data[cols_orig].copy()
    diff_df -= data[cols_shuffled].values

    if rename:
        diff_df.columns = cols_diff
    else:
        diff_df.columns = cols_orig

    return diff_df

    
    # np.random.seed(random_state)
    # data_values = data.values
    # shuffled_indices = np.random.permutation(data_values.shape[1])
    # subtracted_data = data_values - data_values[:, shuffled_indices]

    # # Create new column names based on the subtraction
    # new_column_names = [
    #     f"{data.columns[i]} - {data.columns[shuffled_indices[i]]}"
    #     for i in range(len(data.columns))
    # ]

    # # Create a new DataFrame with the subtracted data and new column names
    # subtracted_df = pd.DataFrame(
    #     subtracted_data, index=data.index, columns=new_column_names
    # )
    # return subtracted_df



#!/usr/bin/env python3
# Time-stamp: "2021-12-18 16:58:11 (ywatanabe)"

import numpy as np
import mngs
import os


class EarlyStopping:
    """Early stops the training if validation score doesn't improve after a given patience."""

    def __init__(
        self, patience=7, verbose=False, delta=0, trace_func=print
    ):
        """
        Args:
            patience (int): How long to wait after last time validation score improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation score improvement.
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
            trace_func (function): trace print function.
                            Default: print
        """
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_score_min = np.Inf
        self.delta = delta

        self.spaths_and_models_dict = {None: None, None: None}

        self.i_epoch = 0
        self.i_global = 0
        self.trace_func = trace_func

    def __call__(
        self, val_score, spaths_and_models_dict, i_epoch, i_global
    ):

        score = -val_score

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(
                val_score, spaths_and_models_dict, i_epoch, i_global
            )

        elif score < self.best_score + self.delta:
            self.counter += 1
            self.trace_func(
                f"\nEarlyStopping counter: {self.counter} out of {self.patience}\n"
            )
            if self.counter >= self.patience:
                self.early_stop = True

        else:
            self.best_score = score
            self.save_checkpoint(
                val_score, spaths_and_models_dict, i_epoch, i_global
            )
            self.counter = 0

    def save_checkpoint(
        self, val_score, spaths_and_models_dict, i_epoch, i_global
    ):
        """Saves model when validation score decrease."""
        if self.verbose:
            self.trace_func(
                f"\nValidation score decreased ({self.val_score_min:.6f} --> "
                f"{val_score:.6f}).  Saving model ...\n"
            )

        for spath, model in spaths_and_models_dict.items():
            try:
                mngs.io.save(model.state_dict(), spath) # torch
            except Exception as e:
                print(e)
                mngs.io.save(model, spath)

        self.i_epoch = i_epoch
        self.i_global = i_global

        ## Update file
        try:
            # rm one-step-old spath
            for spath in list(self.spaths_and_models_dict.keys()):
                os.remove(spath)
                print(f"\nRemoved {spath}\n")
        except Exception as e:
            print(e)

        self.spaths_and_models_dict = spaths_and_models_dict

        self.val_score_min = val_score



#!/usr/bin/env python3
import sys

## https://qiita.com/ynakayama/items/c51c2b6d6a3818bdb63f
# def set_trace():
def ipdb():
    from IPython.core.debugger import Pdb

    # Pdb(color_scheme="Linux").set_trace(sys._getframe().f_back)
    Pdb().set_trace(sys._getframe().f_back)


def debug(f, *args, **kwargs):
    from IPython.core.debugger import Pdb

    # pdb = Pdb(color_scheme="Linux")
    pdb = Pdb()
    return pdb.runcall(f, *args, **kwargs)


################################################################################
## Debugging
################################################################################
def paste():
    from IPython import embed

    embed()


def reload(module_or_func):
    import importlib
    import sys

    try:
        importlib.reload(module_orfunc)
    except:
        pass

    try:
        importlib.reload(sys.modules[module_orfunc.__module__])
    except:
        pass


################################################################################
## Debugging
################################################################################
# https://qiita.com/ynakayama/items/c51c2b6d6a3818bdb63f
# def set_trace():
#     from IPython.core.debugger import Pdb

#     Pdb(color_scheme="Linux").set_trace(sys._getframe().f_back)


# def debug(f, *args, **kwargs):
#     from IPython.core.debugger import Pdb

#     pdb = Pdb(color_scheme="Linux")
#     return pdb.runcall(f, *args, **kwargs)


# def ipdb():
#     import ipdb

#     ipdb.set_trace()


# def reload_func(func):
#     import importlib
#     import sys

#     # importlib.reload(sys.modules['foo'])
#     importlib.reload(sys.modules[func.__module__])
#     # from foo import bar
#     # from sys.modules[func.__module__] import func
#     return func


# def reload_module(module):
#     import importlib

#     module = importlib.reload(module)
#     return module


# def modify_func(func):
#     module = sys.modules[func.__module__]


# def modify_func(fun):
#     modulename = fun.__module__
#     module = __import__(modulename)


# def modify_func(fun):
#     import inspect

#     module = inspect.getmodule(fun)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-27 19:53:28 (ywatanabe)"

import inspect
import os as _os
from datetime import datetime
from glob import glob
from pprint import pprint
from time import sleep

import mngs


def load_configs(is_debug=False):
    CONFIGS = {}
    for lpath in glob("./config/*.yaml"):
        CONFIG = mngs.io.load(lpath)
        # ???
        # fname = mngs.gen.split_fpath(lpath)[1]
        # CONFIGS[fname] = CONFIG
        # Update parameters for debugging
        if is_debug:
            debug_keys = mngs.gen.search("^DEBUG_", list(CONFIG.keys()))[1]
            for dk in debug_keys:
                dk_wo_debug_prefix = dk.split("DEBUG_")[1]
                CONFIG[dk_wo_debug_prefix] = CONFIG[dk]
                print(f"\n{dk} -> {dk_wo_debug_prefix}\n")
            sleep(1)

        CONFIGS.update(CONFIG)

    # # If only one CONFIG files are found, it should be treated as CONFIG's'
    # if len(CONFIGS) == 1:
    #     CONFIGS = CONFIGS[list(CONFIGS.keys())[0]]

    return CONFIGS


def start(
    sys=None,
    plt=None,
    sdir=None,
    show=True,
    # seeds
    os=None,
    random=None,
    np=None,
    torch=None,
    tf=None,
    seed=42,
    # matplotlib
    dpi=100,
    save_dpi=300,
    figsize=(16.2, 10),
    figscale=1.0,
    fontsize=16,
    labelsize="same",
    legendfontsize="xx-small",
    tick_size="auto",
    tick_width="auto",
    hide_spines=False,
):
    """
    import sys
    import matplotlib.pyplot as plt
    import mngs

    CONFIG, sys.stdout, sys.stderr, plt = mngs.gen.start(sys, plt)

    # YOUR CODE HERE

    mngs.gen.close(CONFIG)
    """
    # Timer
    start_time = datetime.now()

    # Debug mode check
    try:
        is_debug = mngs.io.load("./config/is_debug.yaml").get("DEBUG", False)
    except Exception as e:
        is_debug = False

    # ID
    ID = mngs.gen.gen_ID()
    ID = ID if not is_debug else "[DEBUG] " + ID
    print(f"\n{'#'*40}\n## {ID}\n{'#'*40}\n")
    sleep(1)

    # Defines SDIR
    if sdir is None:
        __file__ = inspect.stack()[1].filename
        if "ipython" in __file__:
            __file__ = "/tmp/fake.py"
        spath = __file__
        _sdir, sfname, _ = mngs.general.split_fpath(spath)
        sdir = _sdir + sfname + "/" + ID + "/"  # " # + "/log/"
    _os.makedirs(sdir, exist_ok=True)

    # CONFIGs
    CONFIGS = load_configs(is_debug)
    CONFIGS["ID"] = ID
    CONFIGS["START_TIME"] = start_time
    CONFIGS["SDIR"] = sdir
    if show:
        print(f"\n{'-'*40}\n")
        print(f"CONFIG:")
        for k, v in CONFIGS.items():
            print(f"\n{k}:\n{v}\n")
            sleep(0.3)
        # pprint(CONFIGS)
        print(f"\n{'-'*40}\n")

    # Logging (tee)
    if sys is not None:
        sys.stdout, sys.stderr = mngs.general.tee(sys, sdir=sdir, show=show)

    # Random seeds
    if (
        (os is not None)
        or (random is not None)
        or (np is not None)
        or (torch is not None)
        or (tf is not None)
    ):
        mngs.gen.fix_seeds(
            os=os,
            random=random,
            np=np,
            torch=torch,
            tf=tf,
            seed=seed,
            show=show,
        )

    # Matplotlib configuration
    if plt is not None:
        plt = mngs.plt.configure_mpl(
            plt,
            dpi=dpi,
            save_dpi=save_dpi,
            figsize=figsize,
            figscale=figscale,
            fontsize=fontsize,
            labelsize=labelsize,
            legendfontsize=legendfontsize,
            tick_size=tick_size,
            tick_width=tick_width,
            hide_spines=hide_spines,
            show=show,
        )

    return CONFIGS, sys.stdout, sys.stderr, plt


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import mngs

    # --------------------------------------------------------------------------- #
    CONFIG, sys.stdout, sys.stderr, plt = mngs.gen.start(sys, plt)
    # --------------------------------------------------------------------------- #

    # YOUR CODE HERE
    print("Hello world from mngs.")

    # --------------------------------------------------------------------------- #
    mngs.gen.close(CONFIG)
    # --------------------------------------------------------------------------- #

    # EOF



#!/usr/bin/env python3
# Time-stamp: "2024-01-21 19:17:13 (ywatanabe)"

from functools import partial

import mngs
import numpy as np
import torch
import torch.nn as nn
import torchaudio
from mngs.dsp.HilbertTransformationTorch import HilbertTransformerTorch

# Global definitions
BANDS_LIM_HZ_DICT = {
    "delta": [0.5, 4],
    "theta": [4, 8],
    "lalpha": [8, 10],
    "halpha": [10, 13],
    "beta": [13, 32],
    "gamma": [32, 75],
}


def phase(x, axis=-1):
    x, x_type = mngs.gen.my2tensor(x)

    analytical_x = hilbert(x, axis=axis)
    out = analytical_x[..., 0]

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def phase_band(x, samp_rate, band_str="delta", l=None, h=None):
    """
    Example:
        x = mngs.dsp.demo_sig_torch(x)
        mngs.dsp.phase_band_torch(x, samp_rate=1000, band_str="delta")

        x = mngs.dsp.demo_sig_torch(x)
        mngs.dsp.phase_band_torch(x, samp_rate=1000, band_str=None, l=0.5, h=4)

    Bands definitions:
        BANDS_LIM_HZ_DICT = {
        "delta": [0.5, 4],
        "theta": [4, 8],
        "lalpha": [8, 10],
        "halpha": [10, 13],
        "beta": [13, 32],
        "gamma": [32, 75],
    }
    """
    x, x_type = mngs.gen.my2tensor(x)

    is_band_passed = (band_str is not None) & (l is None) & (h is None)
    is_lh_passed = (band_str is None) & (l is not None) & (h is not None)
    assert is_band_passed or is_lh_passed

    if is_band_passed:
        l, h = BANDS_LIM_HZ_DICT[band_str]

    x = bandpass(x, samp_rate, low_hz=l, high_hz=h)
    out = phase(x)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def amp(x, axis=-1):
    x, x_type = mngs.gen.my2tensor(x)

    analytical_x = hilbert(x, axis=axis)
    out = analytical_x[..., 1]

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def amp_band(x, samp_rate, band_str="delta", l=None, h=None):
    """
    Example:
        x = mngs.dsp.demo_sig_torch(x)
        mngs.dsp.amp_band_torch(x, samp_rate=1000, band_str="delta")

        x = mngs.dsp.demo_sig_torch(x)
        mngs.dsp.amp_band_torch(x, samp_rate=1000, band_str=None, l=0.5, h=4)

    Bands definitions:
        BANDS_LIM_HZ_DICT = {
        "delta": [0.5, 4],
        "theta": [4, 8],
        "lalpha": [8, 10],
        "halpha": [10, 13],
        "beta": [13, 32],
        "gamma": [32, 75],
    }
    """
    x, x_type = mngs.gen.my2tensor(x)

    is_band_passed = (band_str is not None) & (l is None) & (h is None)
    is_lh_passed = (band_str is None) & (l is not None) & (h is not None)
    assert is_band_passed or is_lh_passed

    if is_band_passed:
        l, h = BANDS_LIM_HZ_DICT[band_str]

    x = bandpass(x, samp_rate, low_hz=l, high_hz=h)

    out = amp(x)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def hilbert(x, axis=-1):
    x, x_type = mngs.gen.my2tensor(x)

    if axis == -1:
        axis = x.ndim - 1

    out = HilbertTransformerTorch(axis=axis).to(x.device)(x)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def fft(x, samp_rate, time_dim=-1):
    x, x_type = mngs.gen.my2tensor(x)

    fn = partial(_fft_1d, samp_rate=samp_rate, return_freq=False)
    fft_coef = _apply_to_the_time_dim(fn, x, time_dim)
    freq = torch.fft.fftfreq(x.shape[-1], d=1 / samp_rate)

    if x_type == "numpy":
        return mngs.gen.my2array(fft_coef)[0], mngs.gen.my2array(freq)[0]
    else:
        return fft_coef, freq


def _fft_1d(x, samp_rate, return_freq=True):
    freq = torch.fft.fftfreq(x.shape[-1], d=1 / samp_rate)
    fft_coef = torch.fft.fft(x)  # [:, :, :nyq]

    if return_freq:
        return fft_coef, freq
    else:
        return fft_coef


def rfft_bands(
    x,
    samp_rate,
    bands_str=["delta", "theta", "lalpha", "halpha", "beta", "gamma"],
    normalize=False,
):
    """
    Returns mean absolute rfft coefficients of bands.
    Bands' definitions are as follows.

    BANDS_LIM_HZ_DICT = {
        "delta": [0.5, 4],
        "theta": [4, 8],
        "lalpha": [8, 10],
        "halpha": [10, 13],
        "beta": [13, 32],
        "gamma": [32, 75],
    }

    rfft_bands_p = partial(mngs.dsp.rfft_bands, samp_rate=samp_rate)
    """
    x, x_type = mngs.gen.my2tensor(x)

    coef, freq = rfft(x, samp_rate)
    amp = coef.abs()

    if normalize:
        amp /= amp.sum(axis=-1, keepdims=True)

    amp_bands_abs_mean = []
    for band_str in bands_str:
        low, high = BANDS_LIM_HZ_DICT[band_str]
        indi_band = (low <= freq) & (freq <= high)
        amp_band_abs_mean = amp[..., indi_band].mean(dim=-1)
        amp_bands_abs_mean.append(amp_band_abs_mean)

    out = torch.stack(amp_bands_abs_mean, dim=-1)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def rfft(x, samp_rate, time_dim=-1):
    x, x_type = mngs.gen.my2tensor(x)

    fn = partial(_rfft_1d, samp_rate=samp_rate, return_freq=False)
    fft_coef = _apply_to_the_time_dim(fn, x, time_dim)
    freq = torch.fft.rfftfreq(x.shape[-1], d=1 / samp_rate)

    if x_type == "numpy":
        return mngs.gen.my2array(fft_coef)[0], mngs.gen.my2array(freq)[0]
    else:
        return fft_coef, freq


def _rfft_1d(x, samp_rate, return_freq=True):
    freq = torch.fft.rfftfreq(x.shape[-1], d=1 / samp_rate)
    fft_coef = torch.fft.rfft(x)  # [:, :, :nyq]

    if return_freq:
        return fft_coef, freq
    else:
        return fft_coef


def bandstop(x, samp_rate, low_hz=55, high_hz=65, time_dim=-1):
    x, x_type = mngs.gen.my2tensor(x)

    fn = partial(
        _bandstop_1d, samp_rate=samp_rate, low_hz=low_hz, high_hz=high_hz
    )
    out = _apply_to_the_time_dim(fn, x, time_dim)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def _bandstop_1d(x, samp_rate, low_hz=55, high_hz=65):
    fft_coef, freq = _rfft_1d(x, samp_rate)
    indi_to_cut = (low_hz < freq) & (freq < high_hz)
    fft_coef[indi_to_cut] = 0
    return torch.fft.irfft(fft_coef)


def bandpass(x, samp_rate, low_hz=55, high_hz=65, time_dim=-1):
    x, x_type = mngs.gen.my2tensor(x)

    fn = partial(
        _bandpass_1d, samp_rate=samp_rate, low_hz=low_hz, high_hz=high_hz
    )
    out = _apply_to_the_time_dim(fn, x, time_dim)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def _bandpass_1d(x, samp_rate, low_hz=55, high_hz=65):
    fft_coef, freq = _rfft_1d(x, samp_rate)
    indi_to_cut = (freq < low_hz) + (high_hz < freq)
    fft_coef[indi_to_cut] = 0
    return torch.fft.irfft(fft_coef)


def spectrogram(x, fft_size, device="cuda"):
    """
    Short-time FFT for signals.

    Input: [BS, N_CHS, SEQ_LEN]
    Output: [BS, N_CHS, fft_size//2 + 1, ?]

    spect = spectrogram(x, 50)
    print(spect.shape)
    """
    x, x_type = mngs.gen.my2tensor(x)

    transform = torchaudio.transforms.Spectrogram(n_fft=fft_size).to(device)
    out = transform(x)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def mean(x):
    return x.mean(-1, keepdims=True)


def std(x):
    return x.std(-1, keepdims=True)


def zscore(x):
    x, x_type = mngs.gen.my2tensor(x)

    _mean = mean(x)
    diffs = x - _mean
    var = torch.mean(torch.pow(diffs, 2.0), dim=-1, keepdims=True)
    std = torch.pow(var, 0.5)
    out = diffs / std

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def kurtosis(x):
    x, x_type = mngs.gen.my2tensor(x)

    zscores = zscore(x)
    out = torch.mean(torch.pow(zscores, 4.0), dim=-1, keepdims=True) - 3.0

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def skewness(x):
    x, x_type = mngs.gen.my2tensor(x)

    zscores = zscore(x)
    out = torch.mean(torch.pow(zscores, 3.0), dim=-1, keepdims=True)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def median(x):
    x, x_type = mngs.gen.my2tensor(x)

    out = torch.median(x, dim=-1, keepdims=True)[0]

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def q25(x, q=0.25):
    x, x_type = mngs.gen.my2tensor(x)

    out = torch.quantile(x, q, dim=-1, keepdims=True)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def q75(x, q=0.75):
    x, x_type = mngs.gen.my2tensor(x)

    out = torch.quantile(x, q, dim=-1, keepdims=True)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def rms(x):
    x, x_type = mngs.gen.my2tensor(x)

    out = torch.square(x).sqrt().mean(dim=-1, keepdims=True)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def beyond_r_sigma_ratio(x, r=2.0):
    x, x_type = mngs.gen.my2tensor(x)

    sigma = std(x)
    out = (sigma < x).float().mean(dim=-1, keepdims=True)

    if x_type == "numpy":
        return mngs.gen.my2array(out)[0]
    else:
        return out


def _apply_to_the_time_dim(fn, x, time_dim):
    # Permute the tensor to bring the time dimension to the last position if it's not already
    if time_dim != -1:
        dims = list(range(x.dim()))
        dims[-1], dims[time_dim] = dims[time_dim], dims[-1]
        x = x.permute(*dims)

    # Flatten the tensor along the time dimension
    shape = x.shape
    x = x.reshape(-1, shape[-1])

    # Apply the function to each slice along the time dimension
    applied = torch.stack([fn(x_i) for x_i in torch.unbind(x, dim=0)], dim=0)

    # Reshape the tensor to its original shape (with the time dimension at the end)
    applied = applied.reshape(*shape[:-1], -1)

    # Permute back to the original dimension order if necessary
    if time_dim != -1:
        applied = applied.permute(*dims)

    return applied


# def _apply_to_the_time_dim(fn, x):
#     """
#     x: [BS, N_CHS, SEQ_LEN]
#     When fn(x[0,0]) works, _apply_to_the_time_dim(fn, x) works.
#     """
#     shape = x.shape
#     x = x.reshape(-1, shape[-1])
#     dim = 0
#     applied = torch.stack(
#         [fn(x_i) for x_i in torch.unbind(x, dim=dim)], dim=dim
#     )
#     return applied.reshape(shape[0], shape[1], -1)


# def _test_notch_filter_1d():
#     time = torch.linspace(0, 1, 999)
#     sig = (
#         torch.cos(60 * 2 * torch.pi * time)
#         + torch.cos(200 * 2 * torch.pi * time)
#         + torch.cos(300 * 2 * torch.pi * time)
#     )

#     sig_filted = notch_filter_1d(sig, SAMP_RATE, cutoff_hz=60, width_hz=5)
#     fig, axes = plt.subplots(4, 1)
#     axes[0].plot(sig, label="sig")
#     fft_coef_sig, freq_sig = _rfft_1d(sig, SAMP_RATE)
#     axes[1].plot(freq_sig, fft_coef_sig.abs(), label="fft_coef_sig")
#     axes[2].plot(sig_filted, label="sig_filted")

#     fft_coef_sig_filted, freq_sig_filted = _rfft_1d(sig_filted, SAMP_RATE)
#     axes[3].plot(
#         freq_sig_filted, fft_coef_sig_filted.abs(), label="fft_coef_sig_filted"
#     )
#     for ax in axes:
#         ax.legend()
#     fig.show()


# def test_phase_amp_bandpass():
#     samp_rate = 1000
#     len_seq = 10
#     time = torch.arange(0, len_seq, 1 / samp_rate)
#     freqs_hz = [2, 5, 10]
#     sig = torch.vstack(
#         [torch.sin(f * 2 * torch.pi * time) for f in freqs_hz]
#     ).sum(dim=0)
#     # sig = _bandstop_1d(sig, samp_rate, low_hz=4, high_hz=12)
#     sig = sig.unsqueeze(0).unsqueeze(0).repeat(BS, N_CHS, 1)
#     sig = bandpass(sig, samp_rate, low_hz=0, high_hz=3)

#     phase = _phase_1d(sig)
#     amp = _amp_1d(sig)

#     fig, axes = plt.subplots(3, 1, sharex=True, sharey=True)
#     axes[0].plot(sig[0, 0], label="sig")
#     axes[0].legend()
#     axes[1].plot(phase[0, 0], label="phase")
#     axes[1].legend()
#     axes[2].plot(amp[0, 0], label="amp")
#     axes[2].legend()
#     fig.show()


class FeatureExtractorTorch(nn.Module):
    def __init__(
        self,
        samp_rate,
        features_list=[
            "mean",
            "std",
            "skewness",
            "kurtosis",
            "median",
            "q25",
            "q75",
            "rms",
            "rfft_bands",
            "beyond_r_sigma_ratio",
        ],
        batch_size=8,
    ):
        super().__init__()

        self.func_dict = dict(
            mean=mngs.dsp.mean,
            std=mngs.dsp.std,
            skewness=mngs.dsp.skewness,
            kurtosis=mngs.dsp.kurtosis,
            median=mngs.dsp.median,
            q25=mngs.dsp.q25,
            q75=mngs.dsp.q75,
            rms=mngs.dsp.rms,
            rfft_bands=partial(mngs.dsp.rfft_bands, samp_rate=samp_rate),
            beyond_r_sigma_ratio=mngs.dsp.beyond_r_sigma_ratio,
        )

        self.features_list = features_list

        self.batch_size = batch_size

    def forward(self, x):
        if self.batch_size is None:
            conc = torch.cat(
                [self.func_dict[f_str](x) for f_str in self.features_list],
                dim=-1,
            )

        else:
            conc = []
            n_batches = len(x) // self.batch_size + 1
            for i_batch in range(n_batches):
                try:
                    start = i_batch * self.batch_size
                    end = (i_batch + 1) * self.batch_size
                    conc.append(
                        torch.cat(
                            [
                                self.func_dict[f_str](x[start:end])
                                for f_str in self.features_list
                            ],
                            dim=-1,
                        )
                    )
                except Exception as e:
                    print(e)
            return torch.cat(conc)


def main():
    BS = 32
    N_CHS = 19
    SEQ_LEN = 2000
    SAMP_RATE = 1000
    # NYQ = int(SAMP_RATE / 2)

    x = torch.randn(BS, N_CHS, SEQ_LEN)

    m = FeatureExtractorTorch(SAMP_RATE, batch_size=8)


    out = m(x)  # 15 features

if __name__ == "__main__":
    import mngs

    SAMP_RATE = 1000
    x = mngs.dsp.demo_sig_torch(
        freqs_hz=[2, 3, 5, 10], samp_rate=SAMP_RATE, len_sec=2
    )
    # x = torch.tensor(chirp(time, 3, 500, 100))
    coef, freq = rfft(x, SAMP_RATE)
    amp = coef.abs()



#!/usr/bin/env python3

import csv

import pandas as pd
import mngs
import numpy as np
import scipy

import warnings


if "general" in __file__:
    with warnings.catch_warnings():
        warnings.simplefilter("always")
        warnings.warn(
            '\n"mngs.general.save" will be removed. '
            'Please use "mngs.io.save" instead.',
            PendingDeprecationWarning,
        )


def save(obj, sfname_or_spath, makedirs=True, show=True, **kwargs):
    """
    Example
      save(arr, 'data.npy')
      save(df, 'df.csv')
      save(serializable, 'serializable.pkl')
    """
    import inspect
    import os
    import pickle

    import h5py
    import numpy as np
    import pandas as pd
    import torch
    import yaml

    spath, sfname = None, None

    if "/" in sfname_or_spath:
        spath = sfname_or_spath
    else:
        sfname = sfname_or_spath

    if (spath is None) and (sfname is not None):
        ## for ipython
        __file__ = inspect.stack()[1].filename
        if "ipython" in __file__:
            __file__ = f'/tmp/fake-{os.getenv("USER")}.py'

        ## spath
        fpath = __file__
        fdir, fname, _ = mngs.general.split_fpath(fpath)
        sdir = fdir + fname + "/"
        spath = sdir + sfname
        # spath = mk_spath(sfname, makedirs=True)

    ## Make directory
    if makedirs:
        sdir = os.path.dirname(spath)
        os.makedirs(sdir, exist_ok=True)

    ## Saves
    try:
        ## copy files
        is_copying_files = (isinstance(obj, str) or is_listed_X(obj, str)) and (
            isinstance(spath, str) or is_listed_X(spath, str)
        )
        if is_copying_files:
            mngs.general.copy_files(obj, spath)

        # csv
        elif spath.endswith(".csv"):
            if isinstance(obj, pd.DataFrame):  # DataFrame
                obj.to_csv(spath)
            if is_listed_X(obj, [int, float]):  # listed scalars
                save_listed_scalars_as_csv(
                    obj,
                    spath,
                    **kwargs,
                )
            if is_listed_X(obj, pd.DataFrame):  # listed DataFrame
                # save_listed_dfs_as_csv(obj, spath, indi_suffix=None, overwrite=False)
                save_listed_dfs_as_csv(obj, spath, **kwargs)
        # numpy
        elif spath.endswith(".npy"):
            np.save(spath, obj)
        # pkl
        elif spath.endswith(".pkl"):
            with open(spath, "wb") as s:  # 'w'
                pickle.dump(obj, s)
        # joblib
        elif spath.endswith(".joblib"):
            with open(spath, "wb") as s:  # 'w'
                joblib.dump(obj, s, compress=3)
        # png
        elif spath.endswith(".png"):
            # plotly
            if str(type(obj)) == "<class 'plotly.graph_objs._figure.Figure'>":
                obj.write_image(file=spath, format="png")
            # matplotlib
            else:
                try:
                    obj.savefig(spath)
                except:
                    obj.figure.savefig(spath)
            del obj
        # tiff
        elif spath.endswith(".tiff") or spath.endswith(".tif"):
            obj.savefig(
                spath, dpi=300, format="tiff"
            )  # obj is matplotlib.pyplot object
            del obj
        # mp4
        elif spath.endswith(".mp4"):
            mk_mp4(obj, spath)  # obj is matplotlib.pyplot.figure object
            del obj
        # yaml
        elif spath.endswith(".yaml"):
            with open(spath, "w") as f:
                yaml.dump(obj, f)
        # hdf5
        elif spath.endswith(".hdf5"):
            name_list, obj_list = []
            for k, v in obj.items():
                name_list.append(k)
                obj_list.append(v)
            with h5py.File(spath, "w") as hf:
                for (name, obj) in zip(name_list, obj_list):
                    hf.create_dataset(name, data=obj)
        # pth
        elif spath.endswith(".pth"):
            torch.save(obj, spath)

        # mat
        elif spath.endswith(".mat"):
            scipy.io.savemat(spath, obj)

        # catboost model
        elif spath.endswith(".cbm"):
            obj.save_model(spath)

        else:
            raise ValueError("obj was not saved.")

    except Exception as e:
        print(e)

    else:
        if show and not is_copying_files:
            # if show:
            print(f"\nSaved to: {spath}\n")


def check_encoding(file_path):
    from chardet.universaldetector import UniversalDetector

    detector = UniversalDetector()
    with open(file_path, mode="rb") as f:
        for binary in f:
            detector.feed(binary)
            if detector.done:
                break
    detector.close()
    enc = detector.result["encoding"]
    return enc


def is_listed_X(obj, types):
    """
    Example:
        obj = [3, 2, 1, 5]
        is_listed_X(obj,
    """
    import numpy as np

    try:
        conditions = []
        condition_list = isinstance(obj, list)

        if not (isinstance(types, list) or isinstance(types, tuple)):
            types = [types]

        _conditions_susp = []
        for typ in types:
            _conditions_susp.append(
                (np.array([isinstance(o, typ) for o in obj]) == True).all()
            )

        condition_susp = np.any(_conditions_susp)

        is_listed_X = np.all([condition_list, condition_susp])
        return is_listed_X

    except:
        return False


def save_listed_scalars_as_csv(
    listed_scalars,
    spath_csv,
    column_name="_",
    indi_suffix=None,
    round=3,
    overwrite=False,
    show=False,
):
    """Puts to df and save it as csv"""
    import numpy as np

    if overwrite == True:
        mv_to_tmp(spath_csv, L=2)
    indi_suffix = np.arange(len(listed_scalars)) if indi_suffix is None else indi_suffix
    df = pd.DataFrame(
        {"{}".format(column_name): listed_scalars}, index=indi_suffix
    ).round(round)
    df.to_csv(spath_csv)
    if show:
        print("\nSaved to: {}\n".format(spath_csv))


def save_listed_dfs_as_csv(
    listed_dfs,
    spath_csv,
    indi_suffix=None,
    overwrite=False,
    show=False,
):
    """listed_dfs:
        [df1, df2, df3, ..., dfN]. They will be written vertically in the order.

    spath_csv:
        /hoge/fuga/foo.csv

    indi_suffix:
        At the left top cell on the output csv file, '{}'.format(indi_suffix[i])
        will be added, where i is the index of the df.On the other hand,
        when indi_suffix=None is passed, only '{}'.format(i) will be added.
    """
    import numpy as np

    if overwrite == True:
        mv_to_tmp(spath_csv, L=2)

    indi_suffix = np.arange(len(listed_dfs)) if indi_suffix is None else indi_suffix
    for i, df in enumerate(listed_dfs):
        with open(spath_csv, mode="a") as f:
            f_writer = csv.writer(f)
            i_suffix = indi_suffix[i]
            f_writer.writerow(["{}".format(indi_suffix[i])])
        df.to_csv(spath_csv, mode="a", index=True, header=True)
        with open(spath_csv, mode="a") as f:
            f_writer = csv.writer(f)
            f_writer.writerow([""])
    if show:
        print("Saved to: {}".format(spath_csv))


def mk_mp4(fig, spath_mp4):
    from matplotlib import animation

    axes = fig.get_axes()

    def init():
        return (fig,)

    def animate(i):
        for ax in axes:
            ax.view_init(elev=10.0, azim=i)
        return (fig,)

    anim = animation.FuncAnimation(
        fig, animate, init_func=init, frames=360, interval=20, blit=True
    )

    writermp4 = animation.FFMpegWriter(fps=60, extra_args=["-vcodec", "libx264"])
    anim.save(spath_mp4, writer=writermp4)
    print("\nSaving to: {}\n".format(spath_mp4))


def mv_to_tmp(fpath, L=2):
    import os
    from shutil import move

    try:
        tgt_fname = connect_strs_with_hyphens(fpath.split("/")[-L:])
        tgt_fpath = "/tmp/{}".format(tgt_fname)
        move(fpath, tgt_fpath)
        print("Moved to: {}".format(tgt_fpath))
    except:
        pass


def save_optuna_study_as_csv_and_pngs(study, sdir):
    import optuna

    ## Trials DataFrame
    trials_df = study.trials_dataframe()

    ## Figures
    hparams_keys = list(study.best_params.keys())
    slice_plot = optuna.visualization.plot_slice(study, params=hparams_keys)
    contour_plot = optuna.visualization.plot_contour(study, params=hparams_keys)
    optim_hist_plot = optuna.visualization.plot_optimization_history(study)
    parallel_coord_plot = optuna.visualization.plot_parallel_coordinate(
        study, params=hparams_keys
    )
    hparam_importances_plot = optuna.visualization.plot_param_importances(study)
    figs_dict = dict(
        slice_plot=slice_plot,
        contour_plot=contour_plot,
        optim_hist_plot=optim_hist_plot,
        parallel_coord_plot=parallel_coord_plot,
        hparam_importances_plot=hparam_importances_plot,
    )

    ## Saves
    save(trials_df, sdir + "trials_df.csv")

    for figname, fig in figs_dict.items():
        save(fig, sdir + f"{figname}.png")



#!/usr/bin/env python3


def configure_mpl(
    plt,
    dpi=100,
    save_dpi=300,
    figsize=(16.2, 10),
    figscale=2.0,
    fontsize=10,
    labelsize="same",
    legendfontsize="xx-small",
    tick_size="auto",
    tick_width="auto",
    hide_spines=False,
    show=True,
):
    """
    Configures matplotlib for often used format.

    keyword arguments:
        figsize:
            (width, height) [cm]. 18.1 cm () for 2 columns, 8.7 cm for 1 columns look good for papers.
            Default: (16.2, 10) (= the golden ratio in landscape)

        fontsize:
            Default: 20

        labelsize:
            int (Default: 'same' as the fontsize)

        legendfontsize:
            Default: "xx-small"

        tick_size:
            Major tick size [mm]. 0.8 looks good for papers. (Default: 'auto')

        tick_width:
            Major tick width [mm]. 0.2 looks good for papers. (Default: 'auto')

        hide_spines:
            Hides the top and right spines of the axes. (Default: False)



    Example:
        import matplotlib.pyplot as plt
        import numpy as np

        configure_mpl(plt, figsize=(5, 4), fontsize=7, legendfontsize='small')
        plt.plot(np.random.rand(50))
        plt.show()
    """

    ## shrink the figure a little
    shrink_scale = 0.90

    ## Scales figsize
    figsize_cm = (
        figsize[0] * figscale,
        figsize[1] * figscale,
    )

    figsize_inch = (
        figsize_cm[0] / 2.54,
        figsize_cm[1] / 2.54,
    )

    ## Calculates tick size and width from mm unit.
    dots_on_1_inch_line = dpi**0.5
    dots_on_1_cm_line = dots_on_1_inch_line / 2.54
    dots_on_1_mm_line = dots_on_1_cm_line / 10.0

    if tick_size != "auto":
        tick_size /= dots_on_1_mm_line
    if tick_width != "auto":
        tick_width /= dots_on_1_mm_line

    """
    ## scientific notation
    g = lambda x,pos : "${}$".format(f._formatSciNotation('%1.10e' % x))
    ax.xaxis.set_major_formatter(mticker.FuncFormatter(g))
    ax.yaxis.set_major_formatter(mticker.FuncFormatter(g))

    # plt.ticklabel_format(axis="both", style="plain", scilimits=(-3, 3))


    """

    ## Summarize the props as updater_dict
    updater_dict = {
        "figure.dpi": dpi,
        "savefig.dpi": save_dpi,
        "figure.figsize": (
            figsize_inch[0] * shrink_scale,
            figsize_inch[1] * shrink_scale,
        ),
        "font.size": fontsize,
        "axes.labelsize": fontsize if labelsize == "same" else labelsize,
        "xtick.labelsize": fontsize if labelsize == "same" else labelsize,
        "ytick.labelsize": fontsize if labelsize == "same" else labelsize,
        "axes.titlesize": fontsize if labelsize == "same" else labelsize,
        "legend.fontsize": legendfontsize,
        "pdf.fonttype": 42,
        "ps.fonttype": 42,
        "axes.spines.top": not hide_spines,
        "axes.spines.right": not hide_spines,
        # "figure.autolayout": True,
    }

    ## Tick
    if tick_size != "auto":
        updater_dict["xtick.major.size"] = tick_size
        updater_dict["ytick.major.size"] = tick_size
    if tick_width != "auto":
        updater_dict["xtick.major.width"] = tick_width
        updater_dict["ytick.major.width"] = tick_width
    """
    ## As the example below, number of ticks can be set for each "ax" object.
    max_n_xticks = 4
    max_n_yticks = 4
    for ax in axes:
        ax.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_n_xticks))
        ax.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(max_n_yticks))
    """

    for k, v in updater_dict.items():
        plt.rcParams[k] = v

    updater_dict["figure.figsize"] = str(figsize_cm) + " [cm]"

    if show:
        print(f"\n{'-'*40}")
        print(f"\nMatplotilb has been configured as follows:\n")

        for k, v in updater_dict.items():
            print(k, v)
        print(f"\n{'-'*40}")

    return plt


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    configure_mpl(plt, figsize=(10, 8), fontsize=7, legendfontsize="small")
    plt.plot(np.random.rand(50))
    plt.show()

    # def configure_mpl(
    #     plt, figsize=(1.62, 1), figscale=10, fontsize=20, legendfontsize="xx-small"
    # ):
    #     figbasesize = figsize  # rename for readability

    #     if figsize != (1.62, 1):
    #         figscale = 1

    #     updater_dict = {
    #         # "figure.figsize": (round(1.62 * figscale, 1), round(1 * figscale, 1)),
    #         "figure.figsize": (
    #             round(figsize[0] * figscale, 1),
    #             round(figsize[1] * figscale, 1),
    #         ),
    #         "font.size": fontsize,
    #         "legend.fontsize": legendfontsize,
    #         "pdf.fonttype": 42,
    #         "ps.fonttype": 42,
    #     }
    #     for k, v in updater_dict.items():
    #         plt.rcParams[k] = v
    #     print("\nMatplotilb has been configured as follows:\n{}.\n".format(updater_dict))

    ## EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-20 14:17:34 (ywatanabe)"

import numpy as np
import pandas as pd
import scipy.ndimage
import torch


def gaussian_filter1d(xx, radius):
    """
    Apply a one-dimensional Gaussian filter to an input array, tensor, or DataFrame.

    Arguments:
        xx (numpy.ndarray, torch.Tensor, or pandas.DataFrame): The input data to filter. It can be a 1D or 2D array/tensor/DataFrame.
        radius (int): The radius of the Gaussian kernel. The standard deviation of the Gaussian kernel is implicitly set to 1.

    Returns:
        numpy.ndarray, torch.Tensor, or pandas.DataFrame: The filtered data, with the same type as the input.

    Data Types:
        Input can be either numpy.ndarray, torch.Tensor, or pandas.DataFrame. Output will match the input data type.

    Data Shapes:
        - Input xx: If 1D, shape (n,), if 2D, shape (m, n)
        - Output: Same shape as input xx

    References:
        - SciPy documentation for gaussian_filter1d: https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.gaussian_filter1d.html
    """
    sigma = 1
    truncate = radius / sigma

    # Convert input to NumPy array if it is a pandas DataFrame or PyTorch tensor
    if isinstance(xx, pd.DataFrame):
        values = xx.values
        result = scipy.ndimage.gaussian_filter1d(
            values, sigma, truncate=truncate
        )
        return pd.DataFrame(result, index=xx.index, columns=xx.columns)
    elif isinstance(xx, torch.Tensor):
        values = xx.numpy()
        result = scipy.ndimage.gaussian_filter1d(
            values, sigma, truncate=truncate
        )
        return torch.from_numpy(result)
    else:
        # Assume input is a NumPy array
        return scipy.ndimage.gaussian_filter1d(xx, sigma, truncate=truncate)


def down_sample_1d(x, src_fs, tgt_fs):
    factor = int(src_fs / tgt_fs)
    assert factor == int(factor)
    return scipy.signal.decimate(x, factor)


#!/usr/bin/env python


import time

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.signal
import torch
import torch.nn as nn
import torch.nn.functional as F
from obspy.imaging.cm import obspy_sequential
from obspy.signal.tf_misfit import cwt


class BandPassFilterTorch(nn.Module):
    """
    A PyTorch module that applies a bandpass filter to an input signal using FIR filter coefficients.

    Attributes:
        fs (int): The sampling frequency of the input signal.
        order (int): The order of the filter. Defaults to the sampling frequency if not provided.
        numtaps (int): The number of taps in the filter (order + 1).
        filters (torch.Tensor): The FIR filter coefficients, shaped as (1, 1, numtaps).

    Parameters:
        order (int, optional): The order of the filter. Defaults to the sampling frequency if not provided.
        low_hz (float): The lower cutoff frequency for the bandpass filter.
        high_hz (float): The higher cutoff frequency for the bandpass filter.
        fs (int): The sampling frequency of the input signal.
        n_chs (int): The number of channels in the input signal.

    Methods:
        forward(x): Applies the bandpass filter to the input signal `x`.

    Args:
        x (torch.Tensor): The input signal to be filtered, shaped as (batch_size, n_chs, signal_length) for 3D input
                          or (batch_size, signal_length) for 2D input.

    Returns:
        torch.Tensor: The filtered signal, with the same shape as the input signal.

    References:
        - SincNet implementation: https://raw.githubusercontent.com/mravanelli/SincNet/master/dnn_models.py
        - FIR filter design: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.firwin.html
    """

    def __init__(self, order=None, low_hz=30, high_hz=60, fs=250, n_chs=19):
        super().__init__()
        self.fs = fs
        # nyq = fs / 2
        self.order = fs if order is None else order
        self.numtaps = self.order + 1
        filter_npy = scipy.signal.firwin(
            self.numtaps,
            [low_hz, high_hz],
            pass_zero="bandpass",
            fs=fs,
        )

        self.register_buffer(
            "filters", torch.tensor(filter_npy).unsqueeze(0).unsqueeze(0)
        )

    def forward(self, x):
        """
            Apply the bandpass filter to the input signal.

        The method applies a bandpass filter to the input signal using a FIR filter with the coefficients stored in the `filters` attribute. The filtering is performed in two passes to ensure zero-phase distortion: first forward and then backward.

        Arguments:
            x (torch.Tensor): The input signal tensor. If the input is 3D, it should have the shape (batch_size, n_chs, signal_length), where `n_chs` is the number of channels and `signal_length` is the length of the signal. If the input is 2D, it should have the shape (batch_size, signal_length).

        Returns:
            torch.Tensor: The filtered signal tensor. The output tensor will have the same shape as the input tensor.

        Data Types:
            Input: torch.Tensor (float32 or float64 typically, depending on the model's precision)
            Output: torch.Tensor (same type as input)

        Data Shapes:
            Input: (batch_size, n_chs, signal_length) for 3D input or (batch_size, signal_length) for 2D input
            Output: Same as input

        References:
            - Convolution operation: https://pytorch.org/docs/stable/generated/torch.nn.functional.conv1d.html
            - Zero-phase filtering technique: https://en.wikipedia.org/wiki/Zero-phase_filtering
        """

        dim = x.ndim
        # sig_len_orig = x.shape[-1]

        if dim == 3:
            bs, n_chs, sig_len = x.shape
            x = x.reshape(bs * n_chs, 1, sig_len)

        filted = F.conv1d(
            x, self.filters.type_as(x), padding=int(self.numtaps / 2)
        )
        filted = filted.flip(dims=[-1])  # to backward
        filted = F.conv1d(
            filted, self.filters.type_as(x), padding=int(self.numtaps / 2)
        )
        filted = filted.flip(dims=[-1])  # reverse to the original order

        filted = filted[..., 1:-1]
        # print(self.order, filted.shape[-1])

        if dim == 3:
            filted = filted.reshape(bs, n_chs, -1)

        return filted


class BandPasserCPUTorch:
    """
    A class that applies a bandpass filter to an input signal using a Butterworth filter on the CPU.

    Attributes:
        sos (numpy.ndarray): The second-order sections representation of the filter.

    Parameters:
        low_hz (float): The lower cutoff frequency for the bandpass filter.
        high_hz (float): The higher cutoff frequency for the bandpass filter.
        fs (int): The sampling frequency of the input signal.

    References:
        - Butterworth filter: https://en.wikipedia.org/wiki/Butterworth_filter
        - scipy.signal.butter: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html
        - scipy.signal.sosfilt: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.sosfilt.html
    """

    def __init__(self, low_hz=100, high_hz=250, fs=1000):
        from scipy.signal import butter, sosfilt

        self.butter = butter
        self.sosfilt = sosfilt

        self.sos = self.mk_sos(low_hz, high_hz, fs)

    def __call__(self, raw_sig):
        """
        Apply the bandpass filter to the input signal.

        Parameters:
            raw_sig (numpy.ndarray): The raw input signal. The shape of the array should be (n_samples,) for a 1D signal or (n_channels, n_samples) for a multi-channel signal.

        Returns:
            numpy.ndarray: The filtered signal. The shape of the array will be the same as the input signal.

        Data Types:
            Input: numpy.ndarray (typically float32 or float64)
            Output: numpy.ndarray (same type as input)

        Data Shapes:
            Input: (n_samples,) or (n_channels, n_samples)
            Output: Same as input

        References:
            - scipy.signal.sosfilt: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.sosfilt.html
        """
        filted = self.sosfilt(self.sos, raw_sig)
        return filted

    def mk_sos(self, low_hz, high_hz, fs, order=5):
        """
        Create the second-order sections representation of the Butterworth filter.

        Parameters:
            low_hz (float): The lower cutoff frequency for the bandpass filter.
            high_hz (float): The higher cutoff frequency for the bandpass filter.
            fs (int): The sampling frequency of the input signal.
            order (int): The order of the filter.

        Returns:
            numpy.ndarray: The second-order sections representation of the filter.

        Data Types:
            Return: numpy.ndarray (typically float64)

        Data Shapes:
            Return: (n_sections, 6) where n_sections is the number of second-order filter sections.

        References:
            - scipy.signal.butter: https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.butter.html
        """
        nyq = fs / 2.0
        sos = self.butter(
            order,
            [low_hz / nyq, high_hz / nyq],
            analog=False,
            btype="band",
            output="sos",
        )
        return sos


# from here, not important


class TimeKeeper:
    """
    A simple class to keep track of elapsed time.

    Methods:
        __call__(message=None): Print the elapsed time with an optional message.
        start(): Reset the start time to the current time.
    """

    def __init__(
        self,
    ):
        """
        Initialize the TimeKeeper with the current time as the start time.
        """

        self.start_time = time.time()

    def __call__(self, message=None):
        """
        Print the elapsed time since the start time with an optional message.

        Parameters:
            message (str, optional): An optional message to print before the elapsed time.
        """

        self.elapsed = time.time() - self.start_time
        print(message)
        print("elapsed time: {:.5f} [sec]".format(self.elapsed))

    def start(
        self,
    ):
        """
        Reset the start time to the current time.
        """

        self.start_time = time.time()


if __name__ == "__main__":
    # Parameters
    PASS_LOW_HZ = 40
    PASS_HIGH_HZ = 80
    T_sec = 50.0
    fs = 250
    freqs_hz = [30, 60, 100, 200, 1000]
    n_chs = 19
    bs = 64
    i_batch_show = 0
    i_ch_show = 0
    sig_orig = np.array(
        [
            mngs.dsp.prepare_demo_data(
                T_sec=T_sec, fs=fs, freqs_hz=freqs_hz, n_chs=n_chs
            )
            for _ in range(bs)
        ]
    )
    print("sig_orig_shape: {}".format(sig_orig.shape))
    tk = TimeKeeper()

    # Original
    title_orig = "Original, #{}, ch{} (Demo Freqs: {} Hz)".format(
        i_batch_show, i_ch_show, freqs_hz
    )
    wv_out_orig = wavelet_np(
        sig_orig[i_batch_show, i_ch_show],
        fs,
        f_min=1,
        plot=True,
        title=title_orig,
    )

    # CPU Bandpass Filtering
    bp_cpu = BandPasserCPUTorch(
        low_hz=PASS_LOW_HZ, high_hz=PASS_HIGH_HZ, fs=fs
    )
    filted_cpu = sig_orig.copy()
    ## calculation start ###
    tk.start()
    for i_batch in range(bs):
        for i_ch in range(n_chs):
            filted_cpu[i_batch, i_ch] = bp_cpu(filted_cpu[i_batch, i_ch])
            tk(message="CPU")
            ## calculation end ###
    title_filt_cpu = "[CPU] Bandpass Filted, #{}, ch{} (Freqs: (Low_lim, High_lim) = ({}, {}) Hz) (time: {:.5f} [sec])".format(
        i_batch_show, i_ch_show, PASS_LOW_HZ, PASS_HIGH_HZ, tk.elapsed
    )
    _wv_out_filted_cpu = wavelet_np(
        filted_cpu[i_batch_show, i_ch_show],
        fs,
        f_min=1,
        plot=True,
        title=title_filt_cpu,
    )

    # GPU Bandpass Filtering
    BandPassFilterGPUTorch = BandPassFilterTorch
    bp_gpu = BandPassFilterGPUTorch(
        low_hz=PASS_LOW_HZ, high_hz=PASS_HIGH_HZ, fs=fs
    ).cuda()
    # sig_torch = torch.tensor(sig_orig).unsqueeze(0).unsqueeze(0).cuda()
    sig_torch = torch.tensor(sig_orig).cuda()
    ## calculation start ###
    tk.start()
    filted_gpu = bp_gpu(sig_torch)
    tk(message="GPU")
    ## calculation end ###
    title_filt_gpu = "[GPU] Bandpass Filted, #{}, ch{} (Freqs: (Low_lim, High_lim) = ({}, {}) Hz) (time: {:.5f} [sec])".format(
        i_batch_show, i_ch_show, PASS_LOW_HZ, PASS_HIGH_HZ, tk.elapsed
    )
    _wv_out_filted_gpu = wavelet(
        filted_gpu[i_batch_show, i_ch_show].cpu(),
        fs,
        f_min=1,
        plot=True,
        title=title_filt_gpu,
    )


from scipy.signal import butter, sosfilt, sosfreqz


def bandpassfilter_np(data, lo_hz, hi_hz, fs, order=5):
    """
    Apply a Butterworth bandpass filter to a given data array.

    Parameters:
        data (ndarray): The input signal to be filtered.
        lo_hz (float): The lower cutoff frequency of the bandpass filter in Hz.
        hi_hz (float): The upper cutoff frequency of the bandpass filter in Hz.
        fs (float): The sampling rate of the signal in Hz.
        order (int): The order of the filter. Higher order means a sharper frequency cutoff,
                     but the filter might become unstable. Defaults to 5.

    Returns:
        ndarray: The filtered signal.
    """

    def _mk_butter_bandpass(order=5):
        """
        Create a Butterworth bandpass filter using second-order sections representation.

        Parameters:
            order (int): The order of the filter.

        Returns:
            ndarray: Second-order sections representation of the Butterworth bandpass filter.
        """

        nyq = 0.5 * fs
        low, high = lo_hz / nyq, hi_hz / nyq
        sos = butter(
            order, [low, high], analog=False, btype="band", output="sos"
        )
        return sos

    def _butter_bandpass_filter(data):
        """
        Apply a Butterworth bandpass filter to a given data array.

        Parameters:
            data (ndarray): The input signal to be filtered.

        Returns:
            ndarray: The filtered signal.
        """

        sos = _mk_butter_bandpass()
        y = sosfilt(sos, data)
        return y

    sos = _mk_butter_bandpass(order=order)
    y = _butter_bandpass_filter(data)

    return y

    # EOF


# import scipy


# def gaussian_filter1d(xx, radius):
#     # radius = round(truncate * sigma)
#     sigma = 1
#     truncate = radius / sigma
#     return scipy.ndimage.gaussian_filter1d(xx, sigma, truncate=truncate)



#!/usr/bin/env python3
# Time-stamp: "2021-11-30 11:11:01 (ylab)"

from time import sleep
from pprint import pprint


def get_params(model, tgt_name=None, sleep_sec=2, show=False):

    name_shape_dict = {}
    for name, param in model.named_parameters():
        if (tgt_name is not None) & (name == tgt_name):
            return param
        if tgt_name is None:
            # print(f"\n{param}\n{param.shape}\nname: {name}\n")
            if show is True:
                print(f"\n{param}: {param.shape}\nname: {name}\n")
                sleep(sleep_sec)
            name_shape_dict[name] = list(param.shape)

    if tgt_name is None:
        print()
        pprint(name_shape_dict)
        print()



#!/usr/bin/env python3

import json
import os
import pickle
import warnings

import h5py
import joblib  # [REVISED]
import mne
import numpy as np
import pandas as pd
import torch
import yaml

if "general" in __file__:
    with warnings.catch_warnings():
        warnings.simplefilter("always")
        warnings.warn(
            '\n"mngs.general.load" will be removed. '
            'Please use "mngs.io.load" instead.',
            PendingDeprecationWarning,
        )


def load(lpath, show=False, **kwargs):
    extension = "." + lpath.split(".")[-1]  # [REVISED]

    # csv
    if extension == ".csv":
        obj = pd.read_csv(lpath, **kwargs)
        obj = obj.loc[:, ~obj.columns.str.contains("^Unnamed")]  # [REVISED]
    # tsv
    elif extension == ".tsv":
        obj = pd.read_csv(lpath, sep="\t", **kwargs)

    # excel
    elif extension in [".xls", ".xlsx", ".xlsm", ".xlsb"]:  # [REVISED]
        obj = pd.read_excel(lpath, **kwargs)

    # numpy
    elif extension == ".npy":
        obj = np.load(lpath, allow_pickle=True)  # [REVISED]
    # pkl
    elif extension == ".pkl":
        with open(lpath, "rb") as l:
            obj = pickle.load(l)
    # joblib
    elif extension == ".joblib":
        with open(lpath, "rb") as l:
            obj = joblib.load(l)
    # hdf5
    elif extension == ".hdf5":
        obj = {}
        with h5py.File(lpath, "r") as hf:
            for name in hf:  # [REVISED]
                obj[name] = hf[name][:]
    # json
    elif extension == ".json":
        with open(lpath, "r") as f:
            obj = json.load(f)
    # png
    elif extension == ".png":
        pass
    # tiff
    elif extension in [".tiff", ".tif"]:
        pass
    # yaml
    elif extension == ".yaml":
        with open(lpath) as f:
            obj = yaml.safe_load(f)  # [REVISED]
    # txt
    elif extension in [".txt", ".log"]:
        with open(lpath, "r") as f:  # [REVISED]
            obj = f.read().splitlines()  # [REVISED]
    # pth
    elif extension in [".pth", ".pt"]:
        obj = torch.load(lpath)

    # mat
    elif extension == ".mat":  # [REVISED]
        from pymatreader import read_mat  # [REVISED]

        obj = read_mat(lpath)  # [REVISED]
    # xml
    elif extension == ".xml":  # [REVISED]
        from xml2dict import xml2dict  # [REVISED]

        obj = xml2dict(lpath)  # [REVISED]
    # # edf
    # elif extension == ".edf":  # [REVISED]
    #     obj = mne.io.read_raw_edf(lpath, preload=True)  # [REVISED]
    # con
    elif extension == ".con":  # [REVISED]
        obj = mne.io.read_raw_fif(lpath, preload=True)  # [REVISED]
        obj = obj.to_data_frame()  # [REVISED]
        obj["samp_rate"] = obj.info["sfreq"]  # [REVISED]
    # # mrk
    # elif extension == ".mrk":  # [REVISED]
    #     obj = mne.io.read_mrk(lpath)  # [REVISED]

    # catboost model
    elif extension == ".cbm":  # [REVISED]
        from catboost import CatBoostModel  # [REVISED]

        obj = CatBoostModel.load_model(lpath)  # [REVISED]

    # EEG data
    elif extension in [
        ".vhdr",
        ".vmrk",
        ".edf",
        ".bdf",
        ".gdf",
        ".cnt",
        ".egi",
        ".eeg",
        ".set",
    ]:
        obj = load_eeg_data(lpath, **kwargs)

    else:
        print(f"\nNot loaded from: {lpath}\n")
        return None

    if show:
        print(f"\nLoaded from: {lpath}\n")

    return obj


def load_eeg_data(filename, **kwargs):
    """
    Load EEG data based on file extension and associated files using MNE-Python.

    Parameters:
    filename (str: The path to the file to be loaded.

    Returns:
    raw (mne.io.Raw: The loaded raw EEG data.
    """
    # Get the file extension
    extension = filename.split(".")[-1]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)

        # Load the data based on the file extension
        if extension in ["vhdr", "vmrk"]:
            # Load BrainVision data
            raw = mne.io.read_raw_brainvision(filename, preload=True, **kwargs)
        elif extension == "edf":
            # Load European data format
            raw = mne.io.read_raw_edf(filename, preload=True, **kwargs)
        elif extension == "bdf":
            # Load BioSemi data format
            raw = mne.io.read_raw_bdf(filename, preload=True, **kwargs)
        elif extension == "gdf":
            # Load General data format
            raw = mne.io.read_raw_gdf(filename, preload=True, **kwargs)
        elif extension == "cnt":
            # Load Neuroscan CNT data
            raw = mne.io.read_raw_cnt(filename, preload=True, **kwargs)
        elif extension == "egi":
            # Load EGI simple binary data
            raw = mne.io.read_raw_egi(filename, preload=True, **kwargs)
        elif extension == "set":
            # ???
            raw = mne.io.read_raw(filename, preload=True, **kwargs)
        elif extension == "eeg":
            is_BrainVision = any(
                os.path.isfile(filename.replace(".eeg", ext))
                for ext in [".vhdr", ".vmrk"]
            )
            is_NihonKoden = any(
                os.path.isfile(filename.replace(".eeg", ext))
                for ext in [".21e", ".pnt", ".log"]
            )

            # Brain Vision
            if is_BrainVision:
                filename_v = filename.replace(".eeg", ".vhdr")
                raw = mne.io.read_raw_brainvision(
                    filename_v, preload=True, **kwargs
                )
            # Nihon Koden
            if is_NihonKoden:
                # raw = mne.io.read_raw_nihon(filename, preload=True, **kwargs)
                raw = mne.io.read_raw(filename, preload=True, **kwargs)
        else:
            raise ValueError(f"Unsupported file extension: {extension}")

        return raw


def check_encoding(file_path):
    from chardet.universaldetector import UniversalDetector

    detector = UniversalDetector()
    with open(file_path, mode="rb") as f:
        for binary in f:
            detector.feed(binary)
            if detector.done:
                break
    detector.close()
    enc = detector.result["encoding"]
    return enc


def get_data_path_from_a_package(package_str, resource):
    import importlib
    import os
    import sys

    spec = importlib.util.find_spec(package_str)
    data_dir = os.path.join(spec.origin.split("src")[0], "data")
    resource_path = os.path.join(data_dir, resource)
    return resource_path


def load_yaml_as_an_optuna_dict(fpath_yaml, trial):
    _d = load(fpath_yaml)

    for k, v in _d.items():

        dist = v["distribution"]

        if dist == "categorical":
            _d[k] = trial.suggest_categorical(k, v["values"])

        elif dist == "uniform":
            _d[k] = trial.suggest_int(k, float(v["min"]), float(v["max"]))

        elif dist == "loguniform":
            _d[k] = trial.suggest_loguniform(
                k, float(v["min"]), float(v["max"])
            )

        elif dist == "intloguniform":
            _d[k] = trial.suggest_int(
                k, float(v["min"]), float(v["max"]), log=True
            )

    return _d


def load_study_rdb(study_name, rdb_raw_bytes_url):
    """
    study = load_study_rdb(
        study_name="YOUR_STUDY_NAME",
        rdb_raw_bytes_url="sqlite:///*.db"
    )
    """
    import optuna

    # rdb_raw_bytes_url = "sqlite:////tmp/fake/ywatanabe/_MicroNN_WindowSize-1.0-sec_MaxEpochs_100_2021-1216-1844/optuna_study_test_file#0.db"
    storage = optuna.storages.RDBStorage(url=rdb_raw_bytes_url)
    study = optuna.load_study(study_name=study_name, storage=storage)
    print(f"\nLoaded: {rdb_raw_bytes_url}\n")
    return study



#!/usr/bin/env python3

import inspect
import os

import mngs
import warnings
from glob import glob


if "general" in __file__:
    with warnings.catch_warnings():
        warnings.simplefilter("always")
        warnings.warn(
            '\n"mngs.general.path" will be removed. '
            'Please use "mngs.io.path" instead.',
            PendingDeprecationWarning,
        )


################################################################################
## PATH
################################################################################
def get_this_fpath(when_ipython="/tmp/fake.py"):
    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = when_ipython  # "/tmp/fake.py"
    return __file__


def mk_spath(sfname, makedirs=False):

    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = f'/tmp/fake-{os.getenv("USER")}.py'

    # __file__ = get_current_file_name()

    ## spath
    fpath = __file__
    fdir, fname, _ = split_fpath(fpath)
    sdir = fdir + fname + "/"
    spath = sdir + sfname

    if makedirs:
        os.makedirs(mngs.general.split_fpath(spath)[0], exist_ok=True)

    return spath


def find_the_git_root_dir():
    import git

    repo = git.Repo(".", search_parent_directories=True)
    return repo.working_tree_dir


def split_fpath(fpath):
    """Split a file path to (1) the directory path, (2) the file name, and (3) the file extention
    Example:
        dirname, fname, ext = split_fpath('../data/01/day1/split_octave/2kHz_mat/tt8-2.mat')
        print(dirname) # '../data/01/day1/split_octave/2kHz_mat/'
        print(fname) # 'tt8-2'
        print(ext) # '.mat'
    """
    dirname = os.path.dirname(fpath) + "/"
    base = os.path.basename(fpath)
    fname, ext = os.path.splitext(base)
    return dirname, fname, ext

def touch(fpath):
    import pathlib
    return pathlib.Path(fpath).touch()
    
def find(directory, pattern):
    search_pattern = os.path.join(directory, '**', pattern)
    return glob(search_pattern, recursive=True)
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-25 15:41:31 (ywatanabe)"

import mngs
import numpy as np
import torch


def to_z(x, axis):
    if isinstance(x, torch.Tensor):  # [REVISED]
        return (x - x.mean(dim=axis, keepdim=True)) / x.std(
            dim=axis, keepdim=True
        )  # [REVISED]
    if isinstance(x, np.ndarray):  # [REVISED]
        return (x - x.mean(axis=axis, keepdims=True)) / x.std(
            axis=axis, keepdims=True
        )


# def to_z(x, axis):
#     if torch.Tensor:
#         return (x - x.mean(dim=axis, keepdims=True)) / x.std(dim=axis, keepdims=True)
#     if np.ndarray:
#         return (x - x.mean(axis=axis, keepdims=True)) / x.std(axis=axis, keepdims=True)

#     x, x_type = mngs.gen.my2array(x)

#     dtype_orig = x.dtype
#     x = x.astype(np.float64)
#     z = (x - x.mean(axis=axis, keepdims=True)) / x.std(axis=axis, keepdims=True)
#     z = z.astype(dtype_orig)

#     if x_type == "tensor":
#         return torch.tensor(z)
#     if x_type == "numpy":
#         return z



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-05-15 16:48:33 (ywatanabe)"

from ._SpatialAttention import SpatialAttention
# from ._FreqDropout import FreqDropout
from ._MNet_1000 import MNet_1000, MNet_config
from ._ChannelGainChanger import ChannelGainChanger
from ._FreqGainChanger import FreqGainChanger
from ._BNet import BNet, BNet_config
from ._SwapChannels import SwapChannels
from ._DropoutChannels import DropoutChannels
from ._ResNet1D import ResNet1D, ResNetBasicBlock



#!/usr/bin/env python3

import mngs


################################################################################
## Reproducibility
################################################################################
def fix_seeds(
    os=None, random=None, np=None, torch=None, tf=None, seed=42, show=True
):
    os_str = "os" if os is not None else ""
    random_str = "random" if random is not None else ""
    np_str = "np" if np is not None else ""
    torch_str = "torch" if torch is not None else ""
    tf_str = "tf" if tf is not None else ""

    # https://github.com/lucidrains/vit-pytorch/blob/main/examples/cats_and_dogs.ipynb
    if os is not None:
        import os

        os.environ["PYTHONHASHSEED"] = str(seed)

    if random is not None:
        random.seed(seed)

    if np is not None:
        np.random.seed(seed)

    if torch is not None:
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        # torch.use_deterministic_algorithms(True)

    if tf is not None:
        tf.random.set_seed(seed)

    if show:
        print(f"\n{'-'*40}\n")
        print(
            f"Random seeds of the following packages have been fixed as {seed}"
        )
        print(os_str, random_str, np_str, torch_str, tf_str)
        print(f"\n{'-'*40}\n")


def gen_ID(N=8):
    import random
    import string
    from datetime import datetime

    now = datetime.now()
    now_str = now.strftime("%Y-%m-%d-%H-%M")
    # today_str = now.strftime("%Y-%m%d")
    randlst = [
        random.choice(string.ascii_letters + string.digits) for i in range(N)
    ]
    rand_str = "".join(randlst)
    return now_str + "_" + rand_str


def gen_timestamp():
    from datetime import datetime

    now = datetime.now()
    now_str = now.strftime("%Y-%m%d-%H%M")
    return now_str


class Tee(object):
    """Example:
    import sys

    sys.stdout = Tee(sys.stdout, "stdout.txt")
    sys.stderr = Tee(sys.stderr, "stderr.txt")

    print("abc") # stdout
    print(1 / 0) # stderr
    # cat stdout.txt
    # cat stderr.txt
    """

    def __init__(self, sys_stdout_or_stderr, spath):
        self._files = [sys_stdout_or_stderr, open(spath, "w")]

    def __getattr__(self, attr, *args):
        return self._wrap(attr, *args)

    def _wrap(self, attr, *args):
        def g(*a, **kw):
            for f in self._files:
                res = getattr(f, attr, *args)(*a, **kw)
            return res

        return g


def tee(sys, sdir=None, show=True):
    """
    import sys

    sys.stdout, sys.stderr = tee(sys)

    print("abc")  # stdout
    print(1 / 0)  # stderr
    """

    import inspect
    import os

    ####################
    ## Determines sdir
    ####################
    if sdir is None:
        __file__ = inspect.stack()[1].filename
        if "ipython" in __file__:
            __file__ = "/tmp/fake.py"
        spath = __file__

        _sdir, sfname, _ = mngs.general.split_fpath(spath)
        sdir = _sdir + sfname + "/log/"

    os.makedirs(sdir, exist_ok=True)

    spath_stdout = sdir + "stdout.log"
    spath_stderr = sdir + "stderr.log"

    os.makedirs(mngs.general.split_fpath(spath_stdout)[0], exist_ok=True)

    sys_stdout = Tee(sys.stdout, spath_stdout)
    sys_stderr = Tee(sys.stdout, spath_stderr)

    # os.makedirs("/tmp/mngs/", exist_ok=True)
    # spath_stdout_def = "/tmp/mngs/stdout.log"
    # spath_stderr_def = "/tmp/mngs/stderr.log"
    # sys_stdout = Tee(sys.stdout, spath_stdout_def)
    # sys_stderr = Tee(sys.stderr, spath_stderr_def)

    if show:
        print(f"\n{'-'*40}\n")
        print(
            f"\nStandard Output/Error are going to be logged in the followings: \n  - {spath_stdout}\n  - {spath_stderr}\n"
        )
        print(f"\n{'-'*40}\n")
    return sys_stdout, sys_stderr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-20 14:20:50 (ywatanabe)"#!/usr/bin/env python3

import random

import numpy as np
import scipy


def down_sample_1d(x, src_fs, tgt_fs):
    factor = int(src_fs / tgt_fs)
    assert factor == int(factor)
    return scipy.signal.decimate(x, factor)


def random_pts(max_sec, samp_rate, start_sec=None, dur_sec=10):

    start_sec = (
        random.randint(0, max_sec) if (start_sec is None) else start_sec
    )
    end_sec = start_sec + dur_sec

    start_pts = int(start_sec * samp_rate)
    end_pts = int(end_sec * samp_rate)

    x_time_sec = np.linspace(0, int(dur_sec), int(dur_sec) * samp_rate)

    return x_time_sec, start_sec, end_sec, start_pts, end_pts



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-27 19:24:23 (ywatanabe)"

from datetime import datetime, timedelta
from glob import glob
from time import sleep

import mngs


def format_diff_time(diff_time):
    # Get total seconds from the timedelta object
    total_seconds = int(diff_time.total_seconds())

    # Calculate hours, minutes and seconds
    hours = total_seconds // 3600
    minutes = (total_seconds % 3600) // 60
    seconds = total_seconds % 60

    # Format the time difference as a string
    diff_time_str = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
    return diff_time_str


def close(CONFIG, message=":)", show=True):
    try:
        end_time = datetime.now()
        diff_time = format_diff_time(end_time - CONFIG["START_TIME"])
        CONFIG["TimeSpent"] = diff_time
        del CONFIG["START_TIME"]
    except Exception as e:
        print(e)

    mngs.io.save(CONFIG, CONFIG["SDIR"] + "CONFIG.pkl")

    try:
        sleep(3)
        mngs.gen.notify(
            message=f"[DEBUG]\n" + message,
            ID=CONFIG["ID"],
            log_paths=glob(CONFIG["SDIR"] + "*.log"),
            show=show,
        )
    except Exception as e:
        print(e)


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import mngs

    CONFIG, sys.stdout, sys.stderr, plt = mngs.gen.start(sys, plt, show=False)

    ic("aaa")
    ic("bbb")
    ic("ccc")

    close(CONFIG)



#!/usr/bin/env python3

import numpy as np
import pandas as pd
import torch
from scipy import fftpack

# def psd(signals_2d, samp_rate, normalize=True):
#     return calc_fft_amps(signals_2d, samp_rate, normalize=normalize)


def fft_powers(signals_2d, samp_rate, normalize=True):
    """
    Calculate the power spectrum of the FFT (Fast Fourier Transform) for each signal in a 2D array or tensor.

    Arguments:
        signals_2d (numpy.ndarray or torch.Tensor): A 2D array or tensor containing multiple signals,
                                                     where each row represents a signal.
        samp_rate (float): The sampling rate of the signals.
        normalize (bool, optional): If True, normalize the FFT powers by the sum of powers for each signal.
                                    Defaults to True.

    Returns:
        pandas.DataFrame: A DataFrame containing the FFT powers for each signal. The columns represent the
                          frequencies, and the rows represent the individual signals.

    Data Types:
        Input can be either numpy.ndarray or torch.Tensor. Output is always a pandas.DataFrame.

    Data Shapes:
        - signals_2d: (n_signals, signal_length)
        - Output DataFrame: (n_signals, n_frequencies)

    Example:
        sig_len = 1024
        n_sigs = 32
        signals_2d = np.random.rand(n_sigs, sig_len)
        samp_rate = 256
        fft_powers_df = calc_fft_powers(signals_2d, samp_rate)

    References:
        - NumPy documentation for FFT: https://numpy.org/doc/stable/reference/routines.fft.html
        - PyTorch documentation for FFT: https://pytorch.org/docs/stable/fft.html
        - pandas.DataFrame documentation: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html
    """

    return calc_fft_amps_2d(signals_2d, samp_rate, normalize=normalize)


# def fft_amps(signals_2d, samp_rate, normalize=True):
#     """
#     Example:
#         sig_len = 1024
#         n_sigs = 32
#         signals_2d = np.random.rand(n_sigs, sig_len)
#         samp_rate = 256
#         fft_df = calc_fft_amps(signals_2d, samp_rate)
#     """
#     fft_amps = np.abs(fftpack.fft(signals_2d))
#     fft_freqs = np.fft.fftfreq(signals_2d.shape[-1], d=1.0 / samp_rate)
#     mask = fft_freqs >= 0
#     fft_amps, fft_freqs = fft_amps[:, mask], np.round(fft_freqs[mask], 1)

#     if normalize == True:
#         fft_amps = fft_amps / np.sum(fft_amps, axis=1, keepdims=True)
#         # fft_outs[0] / np.sum(np.array(fft_outs[0]), axis=1, keepdims=True)

#     fft_df = pd.DataFrame(data=fft_amps, columns=fft_freqs.astype(str))
#     return fft_df


def fft_amps(signals_2d, samp_rate, normalize=True):
    """
    Calculate the amplitude of the FFT (Fast Fourier Transform) for each signal in a 2D array or tensor.

    Arguments:
        signals_2d (numpy.ndarray or torch.Tensor): A 2D array or tensor containing multiple signals,
                                                     where each row represents a signal.
        samp_rate (float): The sampling rate of the signals.
        normalize (bool, optional): If True, normalize the FFT amplitudes by the sum of amplitudes for each signal.
                                    Defaults to True.

    Returns:
        pandas.DataFrame: A DataFrame containing the FFT amplitudes for each signal. The columns represent the
                          frequencies, and the rows represent the individual signals.

    Data Types:
        Input can be either numpy.ndarray or torch.Tensor. Output is always a pandas.DataFrame.

    Data Shapes:
        - signals_2d: (n_signals, signal_length)
        - Output DataFrame: (n_signals, n_frequencies)

    Example:
        sig_len = 1024
        n_sigs = 32
        signals_2d = np.random.rand(n_sigs, sig_len)
        samp_rate = 256
        fft_df = calc_fft_amps(signals_2d, samp_rate)

    """
    if isinstance(signals_2d, torch.Tensor):
        signals_2d = (
            signals_2d.detach().cpu().numpy()
        )  # Convert to NumPy array if input is a PyTorch tensor

    fft_amps = np.abs(fftpack.fft(signals_2d))
    fft_freqs = np.fft.fftfreq(signals_2d.shape[-1], d=1.0 / samp_rate)
    mask = fft_freqs >= 0
    fft_amps, fft_freqs = fft_amps[:, mask], np.round(fft_freqs[mask], 1)

    if normalize:
        fft_amps = fft_amps / np.sum(fft_amps, axis=1, keepdims=True)

    fft_df = pd.DataFrame(data=fft_amps, columns=fft_freqs.astype(str))
    return fft_df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-01-20 17:22:43 (ywatanabe)"#!/usr/bin/env python3

import numpy as np

import pandas as pd
import torch


class KeepPD:
    def __init__(self):  # [REVISED]
        self.columns = None
        self.index = None
        self.i_called = 0

    def __call__(self, df):  # [REVISED]
        if self.i_called % 2 == 0:  # [REVISED]
            self.columns = df.columns.copy()
            self.index = df.index.copy()
        if self.i_called % 2 == 1:  # [REVISED]
            df.columns = self.columns
            df.index = self.index
        self.i_called += 1  # [REVISED]


def my2array(arr_or_tensor):
    """
    Convert a PyTorch tensor to a NumPy array. If the input is a NumPy array, pandas DataFrame, or Series, return it unchanged.

    Arguments:
        arr_or_tensor (np.ndarray, pd.DataFrame, pd.Series, or torch.Tensor): The input tensor, array, DataFrame, or Series to convert.

    Returns:
        np.ndarray: The input data as a NumPy array.
    """
    if isinstance(arr_or_tensor, torch.Tensor):
        # Convert PyTorch tensor to NumPy array
        return arr_or_tensor.numpy(), "torch"
    elif isinstance(arr_or_tensor, (np.ndarray, pd.DataFrame, pd.Series)):
        # Input is already a NumPy array, pandas DataFrame, or Series, return it as is
        return arr_or_tensor, "numpy"
    else:
        raise TypeError(
            "Input must be a PyTorch tensor, NumPy array, pandas DataFrame, or pandas Series."
        )


def my2tensor(arr_or_tensor):
    """
    Convert a NumPy array, pandas DataFrame, or Series to a PyTorch tensor. If the input is already a tensor, return it unchanged.

    Arguments:
        arr_or_tensor (np.ndarray, pd.DataFrame, pd.Series, or torch.Tensor): The input array, DataFrame, Series, or tensor to convert.

    Returns:
        torch.Tensor: The input data as a PyTorch tensor.
    """
    if isinstance(arr_or_tensor, (np.ndarray, pd.DataFrame, pd.Series)):
        # Convert NumPy array, pandas DataFrame, or Series to PyTorch tensor
        out_tensor = torch.tensor(
            arr_or_tensor.values
            if isinstance(arr_or_tensor, (pd.DataFrame, pd.Series))
            else arr_or_tensor
        )
        return out_tensor, "numpy"
    elif isinstance(arr_or_tensor, torch.Tensor):
        # Input is already a tensor, return it as is
        return arr_or_tensor, "torch"
    else:
        raise TypeError(
            "Input must be a NumPy array, pandas DataFrame, pandas Series, or a PyTorch tensor."
        )


def two2three_dim(arr_or_tensor, axis=0):
    """
    Add an extra dimension to a 2D array or tensor.

    Arguments:
        arr_or_tensor (np.ndarray or torch.Tensor): The input 2D array or tensor.
        axis (int): The axis along which to insert the new dimension.

    Returns:
        torch.Tensor: The input data as a 3D PyTorch tensor.
    """
    if isinstance(arr_or_tensor, (np.ndarray, torch.Tensor)):
        if arr_or_tensor.ndim != 2:
            raise ValueError("Input must be a 2D array or tensor.")
        # Add an extra dimension
        return torch.unsqueeze(arr_or_tensor, dim=axis)
    else:
        raise TypeError("Input must be a NumPy array or a PyTorch tensor.")


def three2two_dim(arr_or_tensor, axis=0):
    """
    Remove a dimension from a 3D array or tensor.

    Arguments:
        arr_or_tensor (np.ndarray or torch.Tensor): The input 3D array or tensor.
        axis (int): The axis along which to remove the dimension.

    Returns:
        torch.Tensor: The input data as a 2D PyTorch tensor.
    """
    if isinstance(arr_or_tensor, (np.ndarray, torch.Tensor)):
        if arr_or_tensor.ndim != 3:
            raise ValueError("Input must be a 3D array or tensor.")
        # Remove the specified dimension
        return torch.squeeze(arr_or_tensor, dim=axis)
    else:
        raise TypeError("Input must be a NumPy array or a PyTorch tensor.")



#!/usr/bin/env python3

# from .wavelet import wavelet
from .demo_sig import demo_sig_np, demo_sig_torch
from .feature_extractors import (
    FeatureExtractorTorch,
    amp,
    amp_band,
    bandpass,
    bandstop,
    beyond_r_sigma_ratio,
    fft,
    hilbert,
    kurtosis,
    mean,
    median,
    phase,
    phase_band,
    q25,
    q75,
    rfft,
    rfft_bands,
    rms,
    skewness,
    spectrogram,
    std,
    zscore,
)
from .fft import fft_amps, fft_powers
from .filters import (
    BandPasserCPUTorch,
    BandPassFilterTorch,
    bandpassfilter_np,
    gaussian_filter1d,
)
from .mne import to_dig_montage
from .PARAMS import BANDS_LIM_HZ_DICT
from .referencing import common_average  # random_reference
from .referencing import subtract_from_random_column
from .sampling import down_sample_1d
from .transform import arr2skdf, crop, psd, to_z, wavelet, wavelet_np
